from plone.memoize.view import memoize
from plone.memoize.view import memoize_contextless
from plone.tiles import Tile
from plone.tiles.interfaces import IPersistentTile
from Products.CMFCore.utils import getToolByName
from urllib import quote_plus
from zope.component import getMultiAdapter
from zope.component.hooks import getSite
from zope.interface import implements
from zope.security import checkPermission
from castle.cms.tiles.views import getTileView

import json
import logging
import traceback


logger = logging.getLogger('castle.cms')


def _one(val):
    if type(val) in (list, set, tuple):
        if len(val) > 0:
            return val[0]
    return val


class BaseTile(Tile):
    global_editable = False
    edit_label = 'Tile'
    edit_permission = 'cmf.ModifyPortalContent'
    wrap = True

    def get_focal_point(self):
        focal = self.data.get('override_focal_point')
        if not focal:
            return
        try:
            focal = json.loads(focal)
            if len(focal) == 2:
                return focal
        except:
            pass

    @property
    def title(self):
        return self.data.get('title', None)

    @property
    def more_link(self):
        return _one(self.data.get('more_link', None))

    @property
    def more_text(self):
        return self.data.get('more_text', None)

    @property
    @memoize
    def site(self):
        return getSite()

    @property
    @memoize
    def catalog(self):
        return getToolByName(self.site, 'portal_catalog')

    @property
    @memoize_contextless
    def utils(self):
        return getMultiAdapter((self.context, self.request),
                               name="castle-utils")

    @property
    def edit_url(self):
        url = '%s/@@edit-tile/%s/%s' % (
            self.context.absolute_url(), self.__name__, self.id or ''
        )
        if IPersistentTile.providedBy(self):
            url += '?' + self.request.environ.get('QUERY_STRING') or ''
        return url

    @property
    def view_url(self):
        url = '%s/@@%s/%s' % (
            self.context.absolute_url(), self.__name__, self.id or ''
        )
        if IPersistentTile.providedBy(self):
            if hasattr(self.request, 'tile_data'):
                qs = []
                for k, v in self.request.tile_data.items():
                    if not v or k == 'X-Tile-Persistent':
                        continue
                    if isinstance(v, list):
                        for item in v:
                            if isinstance(item, dict):
                                for subk, subv in item.items():
                                    qs.append((
                                        '%s.%s:records' % (k, subk),
                                        str(subv)
                                    ))
                            else:
                                qs.append((
                                    '%s:records' % k,
                                    str(subv)
                                ))
                    else:
                        qs.append((k, str(v)))
                qs = '&'.join([k + '=' + quote_plus(v) for k, v in qs])
            else:
                qs = self.request.environ.get('QUERY_STRING') or ''
            url += '?' + qs
        return url

    def __call__(self):
        self.request.response.setHeader('X-Theme-Disabled', '1')
        try:
            res = self.render()
            if self.global_editable and checkPermission('cmf.ModifyPortalContent', self.context):
                # wrap with tile url div
                config = json.dumps({
                    'label': self.edit_label,
                    'url': self.view_url,
                    'editUrl': self.edit_url
                })
                res = """
<div class="castle-tile-wrapper pat-edittile"
     data-pat-edittile='%s'>%s</div>""" % (
                    config, res)
            else:
                if self.wrap:
                    res = '<div class="castle-tile-wrapper">%s</div>' % res
            return '<html><body>' + res + '</body></html>'
        except:
            path = ['']
            if hasattr(self.context, 'getPhysicalPath'):
                path = self.context.getPhysicalPath()
            logger.error(
                'Error rendering tile on context: %s, data: %s,\n%s' % (
                    '/'.join(path),
                    repr(self.data),
                    traceback.format_exc()))
            return """<html><body>
<p class="tileerror">
  We apologize, there was an error rendering this snippet
</p></body></html>"""

    def render(self):
        return self.index()


class DisplayTypeTileMixin(object):
    display_type_name = ''
    display_type_default = 'default'
    display_type_fallback_view = None

    def render_display(self):
        view = getTileView(
            self.context, self.request, self.display_type_name,
            self.data.get('display_type', self.display_type_default) or self.display_type_default,
            default=self.display_type_default)
        if view is None:
            if self.display_type_fallback_view:
                view = self.display_type_fallback_view(self.context, self.request)
            else:
                return '<div>Warning: No display found</div>'

        view.tile = self
        return view()


class BaseImagesTile(BaseTile):
    implements(IPersistentTile)

    def get_image_data_from_brain(self, brain):
        base_url = brain.getURL()
        return {
            'high': '%s/@@images/image/high' % base_url,
            'large': '%s/@@images/image/large' % base_url,
            'medium': '%s/@@images/image/preview' % base_url,
            'thumb': '%s/@@images/image/thumb' % base_url,
            'original': base_url,
            'title': brain.Title,
            'description': brain.Description,
            'link': '%s/view' % base_url
        }

    def get_image_data(self, im):
        base_url = im.absolute_url()
        related = self.get_related(im) or im
        return {
            'high': '%s/@@images/image/high' % base_url,
            'large': '%s/@@images/image/large' % base_url,
            'medium': '%s/@@images/image/preview' % base_url,
            'thumb': '%s/@@images/image/thumb' % base_url,
            'original': base_url,
            'title': im.Title(),
            'description': im.Description(),
            'link': '%s/view' % related.absolute_url()
        }

    def get_images_in_folder(self, brain):
        if brain.portal_type == 'Folder':
            # get contents
            folder = brain.getObject()
            images = folder.getFolderContents()
            results = []
            for image in images:
                if image.portal_type == 'Image':
                    results.append(self.get_image_data_from_brain(image))
                else:
                    obj = image.getObject()
                    if hasattr(obj, 'image') and hasattr(obj.image, 'contentType'):
                        results.append(self.get_image_data(obj))
            return results
        else:
            return [self.get_image_data_from_brain(brain)]

    @property
    def images(self):
        results = []
        brains = list(self.catalog(UID=self.data.get('images', [])))
        # we need to order this since catalog results are not ordered
        for uid in self.data.get('images') or []:
            found = False
            for brain in brains:
                if brain.UID == uid:
                    found = brain
                    break
            if not found:
                continue
            brains.remove(found)
            if found.is_folderish:
                results.extend(self.get_images_in_folder(brain))
            else:
                results.append(self.get_image_data_from_brain(found))
        return results

    def get_related(self, obj):
        try:
            return obj.relatedItems[0]
        except:
            return None



from castle.cms.tiles.base import BaseImagesTile
from castle.cms.tiles.base import DisplayTypeTileMixin
from castle.cms.tiles.views import BaseTileView
from castle.cms.tiles.views import TileViewsSource
from castle.cms.widgets import ImageRelatedItemsFieldWidget
from castle.cms.widgets import PreviewSelectFieldWidget
from plone.autoform import directives as form
from plone.supermodel import model
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from zope import schema


class DefaultView(BaseTileView):
    name = 'default'
    order = 0
    index = ViewPageTemplateFile('templates/gallery/default.pt')
    tile_name = 'gallery'


class GalleryTile(BaseImagesTile, DisplayTypeTileMixin):
    display_type_name = 'gallery'
    display_type_default = 'default'
    display_type_fallback_view = DefaultView

    render = DisplayTypeTileMixin.render_display


class IGalleryTileSchema(model.Schema):

    form.widget('images', ImageRelatedItemsFieldWidget)
    images = schema.List(
        title=u"Images",
        description=u"Select images or folders of images to display in "
                    u"gallery",
        value_type=schema.Choice(
            vocabulary='plone.app.vocabularies.Catalog'
        )
    )

    form.widget('display_type', PreviewSelectFieldWidget,
                tile_name='gallery')
    display_type = schema.Choice(
        title=u"Display Type",
        source=TileViewsSource('gallery'),
        default='default'
    )



from castle.cms.interfaces import IBusinessData
from castle.cms.widgets import MapPointFieldWidget
from plone.app.registry.browser import controlpanel
from z3c.form.browser.checkbox import CheckBoxFieldWidget


class BusinessControlPanelForm(controlpanel.RegistryEditForm):
    schema = IBusinessData
    schema_prefix = 'castle'
    id = "BusinessControlPanel"
    label = u"Business Settings"
    description = ""

    def updateFields(self):
        super(BusinessControlPanelForm, self).updateFields()
        self.fields['business_coordinates'].widgetFactory = MapPointFieldWidget
        self.fields['business_days_of_week'].widgetFactory = CheckBoxFieldWidget


class BusinessControlPanel(controlpanel.ControlPanelFormWrapper):
    form = BusinessControlPanelForm



from cStringIO import StringIO
from plone import api
from plone.app.redirector.interfaces import IRedirectionStorage
from plone.memoize.instance import memoize
from plone.outputfilters.browser.resolveuid import ResolveUIDView as BaseResolveUIDView
from Products.CMFCore.interfaces import ISiteRoot
from Products.CMFCore.utils import getToolByName
from Products.Five.browser import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from Products.statusmessages.interfaces import IStatusMessage
from urlparse import urlparse
from zExceptions import NotFound
from zope.component import getUtility
from zope.component.hooks import setSite
from zope.i18nmessageid import MessageFactory

import csv


_ = MessageFactory('RedirectionTool')


class ResolveUIDView(BaseResolveUIDView):
    """
    a better resolve uid that cares about referrer and checks if the item is trashed
    """

    def __call__(self):
        referrer = self.request.environ.get('HTTP_REFERER')
        if referrer:
            parsed = urlparse(referrer)
            obj = self.context.restrictedTraverse(parsed.path, None)
            if obj is not None:
                self.context = obj
                portal = getToolByName(obj, 'portal_url').getPortalObject()
                setSite(portal)

        url = None
        catalog = api.portal.get_tool('portal_catalog')
        res = catalog.unrestrictedSearchResults(UID=self.uuid, trashed=False)
        if res:
            url = res[0].getURL()

        if not url:
            raise NotFound("The link you followed is broken")

        if self.subpath:
            url = '/'.join([url] + self.subpath)

        if self.request.QUERY_STRING:
            url += '?' + self.request.QUERY_STRING

        self.request.response.redirect(url, status=301)

        return ''


def absolutize_path(path, context=None, is_alias=True):
    """Check whether `path` is a well-formed path from the portal root, and
       make it Zope-root-relative. If `is_alias` (as opposed to "is_target"),
       also make sure the user has the requisite ModifyAliases permissions to
       make an alias from that path. Return a 2-tuple: (absolute redirection path,
       an error message iff something goes wrong and otherwise '').

    Assume relative paths are relative to `context`; reject relative paths if
    `context` is None.

    """
    portal = getUtility(ISiteRoot)
    err = None
    if path is None or path == '':
        err = (is_alias and _(u"You have to enter an alias.")
               or _(u"You have to enter a target."))
    elif '://' in path:
        err = (is_alias and _(u"An alias is a path from the portal root and doesn't include http:// or alike.")  # noqa
               or _(u"Target path must be relative to the portal root and not include http:// or the like."))  # noqa
    else:
        if path.startswith('/'):
            context_path = "/".join(portal.getPhysicalPath())
            path = "%s%s" % (context_path, path)
        else:
            if context is None:
                err = (is_alias and _(u"Alias path must start with a slash.")
                       or _(u"Target path must start with a slash."))
            else:
                context_path = "/".join(context.getPhysicalPath()[:-1])
                path = "%s/%s" % (context_path, path)
        if not err and is_alias:  # noqa XXX should we require Modify Alias permission on the target as well?
            source = path.split('/')
            while len(source):
                obj = portal.unrestrictedTraverse(source, None)
                if obj is None:
                    source = source[:-1]
                else:
                    break
            if obj is None:
                err = _(u"You don't have the permission to set an alias from the location you provided.")  # noqa
            else:
                pass
                # XXX check if there is an existing alias
                # XXX check whether there is an object
    return path, err


class RedirectsView(BrowserView):
    template = ViewPageTemplateFile('templates/manage-redirects.pt')

    def redirects(self):
        storage = getUtility(IRedirectionStorage)
        portal = getUtility(ISiteRoot)
        context_path = "/".join(self.context.getPhysicalPath())
        portal_path = "/".join(portal.getPhysicalPath())
        redirects = storage.redirects(context_path)
        for redirect in redirects:
            path = redirect[len(portal_path):]
            yield {
                'redirect': redirect,
                'path': path,
            }

    def __call__(self):
        storage = getUtility(IRedirectionStorage)
        request = self.request
        form = request.form
        status = IStatusMessage(self.request)
        errors = {}

        if 'form.button.Add' in form:
            redirection, err = absolutize_path(form.get('redirection'), is_alias=True)
            if err:
                errors['redirection'] = err
                status.addStatusMessage(err, type='error')
            else:
                # XXX check if there is an existing alias
                # XXX check whether there is an object
                del form['redirection']
                storage.add(redirection, "/".join(self.context.getPhysicalPath()))
                status.addStatusMessage(_(u"Alias added."), type='info')
        elif 'form.button.Remove' in form:
            redirects = form.get('redirects', ())
            for redirect in redirects:
                storage.remove(redirect)
            if len(redirects) > 1:
                status.addStatusMessage(_(u"Aliases removed."), type='info')
            else:
                status.addStatusMessage(_(u"Alias removed."), type='info')
        elif 'form.button.cancel' in form:
            self.request.response.redirect(self.context.absolute_url())
        return self.template(errors=errors)

    @memoize
    def view_url(self):
        return self.context.absolute_url() + '/@@manage-aliases'


class RedirectsControlPanel(BrowserView):
    template = ViewPageTemplateFile('templates/redirects-controlpanel.pt')

    def __init__(self, context, request):
        super(RedirectsControlPanel, self).__init__(context, request)
        self.errors = []
        # list of tuples: (line_number, absolute_redirection_path, err_msg, target)

    def redirects(self):
        storage = getUtility(IRedirectionStorage)
        portal = getUtility(ISiteRoot)
        portal_path = "/".join(portal.getPhysicalPath())
        portal_path_len = len(portal_path)
        for redirect in storage:
            if redirect.startswith(portal_path):
                path = redirect[portal_path_len:]
            else:
                path = redirect
            redirectto = storage.get(redirect)
            if redirectto.startswith(portal_path):
                redirectto = redirectto[portal_path_len:]
            yield {
                'redirect': redirect,
                'path': path,
                'redirect-to': redirectto,
            }

    def __call__(self):
        storage = getUtility(IRedirectionStorage)
        portal = getUtility(ISiteRoot)
        request = self.request
        form = request.form
        status = IStatusMessage(self.request)

        if 'form.button.Remove' in form:
            redirects = form.get('redirects', ())
            for redirect in redirects:
                storage.remove(redirect)
            if len(redirects) == 0:
                status.addStatusMessage(_(u"No aliases selected for removal."), type='info')
            elif len(redirects) > 1:
                status.addStatusMessage(_(u"Aliases removed."), type='info')
            else:
                status.addStatusMessage(_(u"Alias removed."), type='info')
        elif 'form.button.Upload' in form:
            self.upload(form['file'], portal, storage, status)

        return self.template()

    def upload(self, file, portal, storage, status):
        """Add the redirections from the CSV file `file`. If anything goes wrong, do nothing."""
        # Turn all kinds of newlines into LF ones. The csv module doesn't do
        # its own newline sniffing and requires either \n or \r.
        file = StringIO('\n'.join(file.read().splitlines()))

        # Use first two lines as a representative sample for guessing format,
        # in case one is a bunch of headers.
        dialect = csv.Sniffer().sniff(file.readline() + file.readline())
        file.seek(0)

        successes = []  # list of tuples: (abs_redirection, target)
        had_errors = False
        for i, fields in enumerate(csv.reader(file, dialect)):
            if len(fields) == 2:
                redirection, target = fields
                abs_redirection, err = absolutize_path(redirection, is_alias=True)
                abs_target, target_err = absolutize_path(target, is_alias=False)
                if err and target_err:
                    err = "%s %s" % (err, target_err)  # sloppy w.r.t. i18n
                elif target_err:
                    err = target_err
                else:
                    if abs_redirection == abs_target:
                        # TODO: detect indirect recursion
                        err = _(u"Aliases that point to themselves will cause"
                                u"an endless cycle of redirects.")
            else:
                err = _(u"Each line must have 2 columns.")

            if not err:
                if not had_errors:  # else don't bother
                    successes.append((abs_redirection, abs_target))
            else:
                had_errors = True
                self.errors.append(dict(line_number=i+1, line=dialect.delimiter.join(fields),
                                        message=err))

        if not had_errors:
            for abs_redirection, abs_target in successes:
                storage.add(abs_redirection, abs_target)
            status.addStatusMessage(_(u"%i aliases added.") % len(successes), type='info')

    @memoize
    def view_url(self):
        return self.context.absolute_url() + '/@@aliases-controlpanel'



from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support import expected_conditions as EC  # noqa
from selenium.webdriver.support.ui import WebDriverWait
from zope.component import getMultiAdapter

import castle
import os
import time


src_dir = os.path.sep.join(castle.__file__.split(os.path.sep)[:-3])


def dismiss_tour(driver, wait=1):
    try:
        el = WebDriverWait(driver, wait).until(
            EC.element_to_be_clickable((By.CSS_SELECTOR, '.introjs-skipall')))
        el.click()
    except:
        return
    if el:
        try:
            el.click()
        except:
            pass


def screenshot(selenium, name):
    filepath = os.path.join(src_dir, 'error_clicking_%s.png' % name)
    selenium.get_screenshot_as_file(filepath)
    return filepath


def click_it(driver, method, locator, wait=10):
    try:
        element = WebDriverWait(driver, wait).until(
            EC.element_to_be_clickable((method, locator)))
        element.click()
    except:
        # log error but still raise exception because we don't want to have
        # to wait for the tests to finish in order to inspect the problem
        filepath = screenshot(driver, locator)
        raise Exception('Unable to find or click element {}\n{}'.format(
            locator, filepath))


def clear_it(web_element):
    try:
        web_element.clear()
    except:
        text_length = len(web_element.text)
        while text_length > 0:
            web_element.send_keys(Keys.DELETE)
            text_length -= 1


def get_element(driver, method, locator, wait=10,
                detection=EC.visibility_of_element_located):
    try:
        element = WebDriverWait(driver, wait).until(detection((method, locator)))
        return element
    except:
        filepath = screenshot(driver, locator)
        raise Exception('unable to find element %s\n%s' % (
            locator, filepath
        ))


def mousedown(driver, method, locator):
    # this will simulate a mousedown and up event.  This is a lower
    # level interface with mouse events and in the background is calling a mousedown
    # some of the mosaic toolbar items need this instead of a normal "click"
    action_element = get_element(driver, method, locator)
    actionchain = ActionChains(driver)
    actionchain.click(action_element)
    actionchain.perform()


def create_page(selenium, pagetype='document'):
    click_it(selenium, By.CSS_SELECTOR, '.castle-toolbar-add a')
    click_it(selenium, By.CSS_SELECTOR, '.contenttype-document-container a')
    title_field = get_element(selenium, By.CSS_SELECTOR, '#contentTitle')
    title_field.send_keys('test_' + pagetype)
    click_it(selenium, By.CSS_SELECTOR,
             '#add-modal-react-container .modal-footer .plone-btn-primary')

    tmpcss = pagetype
    if pagetype == 'basic' or pagetype == 'document':
        tmpcss = 'default/' + tmpcss
    else:
        tmpcss = 'castle/' + tmpcss

    # open the select layout dialog
    click_it(selenium, By.CSS_SELECTOR, '.mosaic-button-layout')
    click_it(selenium, By.CSS_SELECTOR, '.mosaic-button-changelayout')
    click_it(selenium, By.CSS_SELECTOR,
             '.mosaic-select-layout [data-value="%s.html"]' % tmpcss)
    click_it(selenium, By.CSS_SELECTOR, '.mosaic-button-save')


def create_news_item(selenium, newsitem_title):
    get_element(selenium, By.CSS_SELECTOR, '.castle-toolbar-add a').click()
    click_it(selenium, By.CSS_SELECTOR, '.contenttype-news-item-container a')
    title_field = get_element(selenium, By.CSS_SELECTOR, '#contentTitle')
    title_field.send_keys('test_' + newsitem_title)
    click_it(selenium, By.CSS_SELECTOR,
             '#add-modal-react-container .modal-footer .plone-btn-primary')
    click_it(selenium, By.CSS_SELECTOR, '.mosaic-button-save')


def create_folder(selenium, foldertype, name=None):
    click_it(selenium, By.CSS_SELECTOR, '.castle-toolbar-add a')
    click_it(selenium, By.CSS_SELECTOR, '.contenttype-folder-container a')
    entry_field = get_element(selenium, By.CSS_SELECTOR, '#contentTitle')
    title = name
    if title is None:
        title = 'test_' + foldertype
    entry_field.send_keys(title)
    click_it(selenium, By.CSS_SELECTOR,
             '#add-modal-react-container .modal-footer .plone-btn-primary')
    click_it(selenium, By.CSS_SELECTOR, '.mosaic-button-layout')
    click_it(selenium, By.CSS_SELECTOR, '.mosaic-button-changelayout')
    click_it(selenium, By.CSS_SELECTOR,
             '.mosaic-select-layout [data-value="castle/%s.html"]' % foldertype)
    click_it(selenium, By.CSS_SELECTOR, '.mosaic-button-save')


def prep_file_upload(selenium, filetoupload, what):
    get_element(selenium, By.CSS_SELECTOR, '.castle-toolbar-add a').click()
    get_element(selenium, By.CSS_SELECTOR, ".autotoc-level-upload").click()
    click_it(selenium, By.CSS_SELECTOR, '.castle-upload-container .droparea button')
    # send file name to hidden input field
    get_element(selenium, By.CSS_SELECTOR,
                '.castle-upload-container input[type="file"]',
                detection=EC.presence_of_element_located).send_keys(filetoupload)
    # enter title
    get_element(
        selenium, By.CSS_SELECTOR,
        '#castle-upload-field-title').click()
    get_element(
        selenium, By.CSS_SELECTOR,
        '#castle-upload-field-title').send_keys(what)
    # enter description
    # for some reason the element needs to be clicked prior to entering text into it.
    get_element(
        selenium, By.CSS_SELECTOR,
        '#castle-upload-field-description').click()
    get_element(
        selenium, By.CSS_SELECTOR,
        '#castle-upload-field-description').send_keys(what)


def add_tag(selenium, tagtoadd):
    # click edit
    click_it(selenium, By.CSS_SELECTOR, '.castle-toolbar-edit a')
    # click properties
    click_it(selenium, By.CSS_SELECTOR, '.mosaic-button-properties')
    # click category tab
    click_it(selenium, By.LINK_TEXT, "Categorization")
    # type tag into tag select2
    entry_field = get_element(
        selenium, By.CSS_SELECTOR, '#formfield-form-widgets-IDublinCore-subjects .select2-input')
    entry_field.send_keys(tagtoadd)
    time.sleep(1)
    # press enter to add tag
    entry_field.send_keys(Keys.ENTER)
    click_it(selenium, By.CSS_SELECTOR, '.mosaic-overlay-ok-button')
    # save
    click_it(selenium, By.CSS_SELECTOR, '.mosaic-button-save')


def get_tile(request, context, name, data):
    tile = getMultiAdapter((context, request), name=name)

    for key in data.iterkeys():
        tile.data[key] = data[key]

    return tile


def render_tile(request, context, name, data):
    tile = get_tile(request, context, name, data)
    page = tile.render()
    return page



from castle.cms import utils
from castle.cms.behaviors.location import ILocation
from castle.cms.interfaces import IFieldTileRenderer
from castle.cms.interfaces import IVersionViewLayer
from datetime import datetime
from plone.app.contenttypes.behaviors.leadimage import ILeadImage
from plone.app.dexterity.behaviors.metadata import IDublinCore
from plone.app.dexterity.behaviors.metadata import IOwnership
from plone.app.dexterity.behaviors.metadata import IPublication
from plone.app.relationfield.behavior import IRelatedItems
from plone.app.standardtiles.field import DexterityFieldTile as BaseDexterityFieldTile
from plone.app.textfield.value import RichTextValue
from plone.autoform.view import WidgetsView
from plone.memoize.view import memoize_contextless
from plone.namedfile.file import NamedBlobImage
from plone.tiles import Tile
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from z3c.form.field import Fields
from zope.component import getMultiAdapter
from zope.component import queryUtility


# optimize lookups
_supported_schemas = {
    'IDublinCore': IDublinCore,
    'IOwnership': IOwnership,
    'IPublication': IPublication,
    'IRelatedItems': IRelatedItems,
    'ILeadImage': ILeadImage,
    'ILocation': ILocation
}

class DexterityFieldTile(BaseDexterityFieldTile):
    default_template = ViewPageTemplateFile('templates/fields/default.pt')
    image_template = ViewPageTemplateFile('templates/fields/image.pt')
    default_timezone = None

    @property
    @memoize_contextless
    def utils(self):
        return getMultiAdapter((self.context, self.request),
                               name="castle-utils")

    def __init__(self, context, request):
        if IVersionViewLayer.providedBy(request):
            version = request.form.get('version')
            context = utils.get_object_version(context, version)
        Tile.__init__(self, context, request)
        WidgetsView.__init__(self, context, request)

        try:
            self.schema_name, _, self.field_name = self.data['field'].partition('-')
        except KeyError:
            self.field = self.field_name = self.schema_name = None
            return

        if self.schema_name in _supported_schemas:
            schema = _supported_schemas[self.schema_name]
            self.field = '%s.%s' % (self.schema_name, self.field_name)
            self.fields += Fields(
                schema,
                prefix=self.schema_name
            ).select(self.field)
            self._additionalSchemata = (schema,)
        else:
            super(DexterityFieldTile, self).__init__(context, request)

    def __call__(self):
        if self.field and self.isVisible:
            renderer = queryUtility(IFieldTileRenderer, self.field + '-renderer')
            if renderer:
                self.field_value = renderer(self)
            else:
                self.update()
                widget = self.widgets.get(self.field)
                if hasattr(widget, '_converter'):
                    self.field_value = widget._converter(
                        widget.field, self).toFieldValue(widget.value)
                else:
                    if widget is None:
                        if hasattr(self.context, self.field):
                            self.field_value = getattr(self.context, self.field)
                        else:
                            self.field_value = ''  # default here...
                    else:
                        self.field_value = widget.value

            if 'className' not in self.data:
                className = ''
                if self.field_name == 'title':
                    className = 'documentFirstHeading'
                elif self.field_name == 'description':
                    className = 'documentDescription'
                self.data['className'] = className
            if 'tag' not in self.data:
                tag = 'div'
                if self.field_name == 'title':
                    tag = 'h1'
                self.data['tag'] = tag

            if isinstance(self.field_value, datetime):
                self.field_value = self.utils.format_date(
                    self.field_value, self.data.get('format', 'notime'))
            elif isinstance(self.field_value, RichTextValue):
                self.field_value = self.field_value.output
            elif isinstance(self.field_value, list):
                self.field_value = '<br />'.join(self.field_value)
            elif isinstance(self.field_value, NamedBlobImage):
                return self.image_template()
            return self.default_template()
        return u'<html></html>'


def LocationsFieldRenderer(view):
    return view.utils.get_locations(view.context)



# -*- coding: utf-8 -*-
from plone.supermodel import model
from plone.tiles import Tile
from zope import schema
from zope.i18nmessageid import MessageFactory
import logging
from castle.cms.fragments.utils import getFragment

_ = MessageFactory('castle.cms')

logger = logging.getLogger('castle.cms')


class IFragmentTile(model.Schema):
    name = schema.TextLine(
        title=_(u'Name'),
    )


class FragmentTile(Tile):
    """A tile that displays a theme fragment"""

    def update(self):
        try:
            self.index = getFragment(self.context, self.request,
                                     self.data['name'].encode('utf-8'))
        except KeyError:
            logger.error(u"Theme fragment '{0:s}' was not found.".format(
                self.data['name']
            ))
            self.index = lambda: u''

    def __call__(self):
        self.update()
        return u'<html><body>{0:s}</body></html>'.format(self.index())



from castle.cms import impersonator
from plone import api
from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from zExceptions import NotFound
from zope.component import getMultiAdapter


class ImpersonatorView(BrowserView):
    index = ViewPageTemplateFile('templates/impersonate.pt')

    def __call__(self):
        resp = self.request.response
        return_url = self.request.form.get('return_url')
        if not return_url:
            return_url = self.context.absolute_url()
        if self.request.form.get('action') == 'stop':
            resp.expireCookie(impersonator.COOKIE_NAME)
            return resp.redirect(return_url)
        elif self.request.REQUEST_METHOD == 'POST':
            authenticator = getMultiAdapter((self.context, self.request),
                                            name=u"authenticator")
            if authenticator.verify():
                resp.setCookie(impersonator.COOKIE_NAME, self.request.form.get('username'))
                return resp.redirect(return_url)

        if api.user.is_anonymous():
            raise NotFound()
        return self.index()



from castle.cms import audit
from castle.cms import tasks
from castle.cms.constants import DEFAULT_SITE_LAYOUT_REGISTRY_KEY
from castle.cms.lead import check_lead_image
from plone import api
from plone.api.exc import CannotGetPortalError
from plone.app.blocks.interfaces import DEFAULT_CONTENT_LAYOUT_REGISTRY_KEY
from plone.app.blocks.layoutbehavior import ILayoutAware
from plone.app.dexterity.behaviors.metadata import IOwnership
from plone.app.dexterity.behaviors.metadata import IPublication
from plone.app.event.base import localized_now
from plone.registry.interfaces import IRegistry
from plone.uuid.interfaces import IUUID
from Products.CMFCore.WorkflowCore import WorkflowException
from Products.CMFPlone.browser.syndication.settings import FeedSettings
from zope.component import getUtility
from zope.globalrequest import getRequest
from zope.interface import Interface


try:
    from z3c.relationfield.interfaces import IRelationBrokenEvent
except ImportError:
    class IRelationBrokenEvent(Interface):
        pass


def onFileEdit(obj, event):
    if IRelationBrokenEvent.providedBy(event):
        # these trigger too much!
        return

    try:
        tasks.file_edited.delay(obj)
    except CannotGetPortalError:
        pass


def onFileDelete(obj, event):
    try:
        tasks.aws_file_deleted.delay(IUUID(obj))
    except CannotGetPortalError:
        pass


def onFileStateChanged(obj, event):
    try:
        tasks.workflow_updated.delay(obj)
    except CannotGetPortalError:
        pass


def onContentCreated(obj, event):
    if obj.portal_type == 'Dashboard':
        return
    metadata = IPublication(obj, None)
    if metadata is not None:
        if metadata.effective is None:
            metadata.effective = localized_now(obj)
    _touchContributors(obj)

    if obj.portal_type == 'Collection':
        # enable syndication on type by default
        settings = FeedSettings(obj)
        settings.enabled = True

    adapted = ILayoutAware(obj, None)
    if adapted:
        if not adapted.content and not adapted.contentLayout:
            registry = getUtility(IRegistry)
            try:
                default_layout = registry['%s.%s' % (
                    DEFAULT_CONTENT_LAYOUT_REGISTRY_KEY,
                    obj.portal_type.replace(' ', '-'))]
                adapted.contentLayout = default_layout
            except (KeyError, AttributeError):
                pass
            try:
                default_layout = registry['%s.%s' % (
                    DEFAULT_SITE_LAYOUT_REGISTRY_KEY,
                    obj.portal_type.replace(' ', '-'))]
                adapted.pageSiteLayout = default_layout
            except (KeyError, AttributeError):
                pass

    try:
        tasks.scan_links.delay('/'.join(obj.getPhysicalPath()))
    except CannotGetPortalError:
        pass

    obj.reindexObject()


def onContentModified(obj, event):
    if IRelationBrokenEvent.providedBy(event):
        # these trigger too much!
        return
    if obj.portal_type == 'Dashboard':
        return
    try:
        tasks.scan_links.delay('/'.join(obj.getPhysicalPath()))
    except CannotGetPortalError:
        pass
    _touchContributors(obj)


def onEditFinished(obj, event):
    """
    on forms submission of done editing page...
    """
    check_lead_image(obj, request=getRequest())


def onObjectEvent(obj, event):
    if IRelationBrokenEvent.providedBy(event):
        # these trigger too much!
        return
    audit.event(obj, event)


def onPASEvent(event):
    audit.event(event)


def _touchContributors(obj):
    ownership = IOwnership(obj, None)
    if ownership is not None:
        # current user id
        try:
            user_id = api.user.get_current().getId()
            if (user_id not in ownership.creators and
                    user_id not in ownership.contributors):
                ownership.contributors = ownership.contributors + (user_id.decode('utf8'),)
        except:
            pass


def onTrashTransitioned(obj, event):
    api.portal.show_message(
        'You are not allowed to transition an item in the recycling bin.',
        request=getRequest(), type='error')
    raise WorkflowException('You are not allowed to transition an item in the recycling bin.')



from castle.cms.files import aws
from castle.cms.interfaces import IReferenceNamedImage
from cStringIO import StringIO
from os import fstat
from PIL import Image
from plone import api
from plone.app.blob.download import handleRequestRange
from plone.app.blob.iterators import BlobStreamIterator
from plone.app.blob.utils import openBlob
from plone.app.imaging.utils import getAllowedSizes
from plone.namedfile import browser as namedfile
from plone.namedfile.interfaces import INamedBlobFile
from plone.namedfile.scaling import ImageScaling
from Products.Five import BrowserView
from webdav.common import rfc1123_date
from zExceptions import NotFound

import base64
import zope.publisher.interfaces


class DownloadAsPNG(BrowserView):
    base64 = False

    def get_data(self):
        if not self.context.image:
            raise NotFound

        blob = self.context.image._blob
        return openBlob(blob)

    def __call__(self):
        fi = self.get_data()
        im = Image.open(fi)

        changed = StringIO()
        im.save(changed, 'PNG')
        changed.seek(0)
        if hasattr(fi, 'close'):
            fi.close()
        data = changed.read()

        if self.base64:
            data = base64.b64encode(data)

        resp = self.request.response
        resp.setHeader('Content-Disposition',
                       'inline; filename=%s.png' % self.context.getId())
        resp.setHeader("Content-Length", data)
        resp.setHeader('Content-Type', 'image/png')
        return data


class ConvertToPNG(DownloadAsPNG):
    base64 = True

    def get_data(self):
        return self.request.form['data']


class Download(namedfile.Download):

    def serve_file(self):
        file = self._getFile()
        self.set_headers(file)
        if not INamedBlobFile.providedBy(file):
            return super(Download, self).__call__()

        request_range = handleRequestRange(
            self.context, file.getSize(), self.request, self.request.response)
        return BlobStreamIterator(file._blob, **request_range)

    def __call__(self):
        if not aws.uploaded(self.context):
            return self.serve_file()

        url = aws.get_url(self.context)
        self.request.response.redirect(url)


class DownloadBlob(BrowserView):

    content_type = 'application/pdf'
    file_ext = 'pdf'

    def get_data(self):
        raise NotImplemented()

    def __call__(self):
        data = self.get_data()
        if data:
            is_blob = False
            if isinstance(data, basestring):
                length = len(data)
            else:
                is_blob = True
                blobfi = openBlob(data)
                length = fstat(blobfi.fileno()).st_size
                blobfi.close()

            self.request.response.setHeader(
                'Last-Modified',
                rfc1123_date(self.context._p_mtime))
            resp = self.request.response
            resp.setHeader('Content-Disposition',
                           'inline; filename=%s.%s' % (self.context.getId(), self.file_ext))
            resp.setHeader("Content-Length", length)
            resp.setHeader('Content-Type', self.content_type)

            if is_blob:
                resp.setHeader('Accept-Ranges', 'bytes')
                range = handleRequestRange(
                    self.context, length, self.request,
                    self.request.response)
                return BlobStreamIterator(data, **range)
            else:
                return data
        else:
            raise NotFound


class EmptyRedirect(BrowserView):
    '''
    I just want to redirect.
    Yes, this doc string is required because Zope2 is insane
    '''

    __roles__ = ('Anonymous',)

    def __init__(self, *args):
        super(EmptyRedirect, self).__init__(*args)
        self.sizes = getAllowedSizes()

    def __call__(self):
        return self

    def __getattr__(self, name):
        if name in self.sizes:
            return self
        raise AttributeError


class CastleImageScaling(ImageScaling):

    def publishTraverse(self, request, name):
        if name == 'image':
            if IReferenceNamedImage.providedBy(self.context.image):
                # auth not setup yet, we just redirect and assume we'll figure the rest out
                catalog = api.portal.get_tool('portal_catalog')
                brains = catalog.unrestrictedSearchResults(UID=self.context.image.reference)
                if len(brains) > 0:
                    brain = brains[0]
                    url = request.ACTUAL_URL.replace(
                        self.context.absolute_url(), brain.getURL())
                    request.response.redirect(url)
                    view = EmptyRedirect(self.context, self.request)
                    return view.__of__(self.context)
        return super(CastleImageScaling, self).publishTraverse(request, name)


class NamedFileDownload(Download):
    def __call__(self):
        try:
            return super(NamedFileDownload, self).__call__()
        except AttributeError:
            raise zope.publisher.interfaces.NotFound(self, self.fieldname, self.request)



from plone.app.textfield import RichTextValue

def getTileData():
    TOP_FULL_WIDTH = u"""
        <div class="text-center">
            <blockquote>Castle CMS powers websites, intranets, and web applications for the most demanding organizations. Built on top of&nbsp;Plone, the&nbsp;dependable open source content management system (CMS), Castle CMS provides a unique combination of world-class&nbsp;content management features and maximum&nbsp;security, and is the result of our team's years of experience leading&nbsp;Plone development&nbsp;and&nbsp;security.</blockquote>
        </div>"""

    THREE_WIDE_1 = u"""
                    <h4 class="text-center">Collaborate Effectively</h4>
                    <p>Castle CMS was built for large team collaboration. Fine-grained security for group editing and viewing,&nbsp;versioning, tagging, and content quality checks: all the features&nbsp;you need to create and manage content worthy of any enterprise.</p>
                    """

    THREE_WIDE_2 = u"""
                    <h4 class="text-center">Theming, Your Way</h4>
                    <p>Create and modify&nbsp;your website theme live, using&nbsp;only a web browser, with the freedom to customize all aspects of Castle CMS. Apply your corporate brand&nbsp;exactly as you envisioned, with the most powerful and flexible theming engine available.</p>
                    """

    THREE_WIDE_3 = u"""
                    <h4 class="text-center">Social Media Integration</h4>
                    <p>Engage your users, customers and audience more effectively than ever before, using Castle CMS' built-in&nbsp;social media platform integrations: Facebook, Twitter, Pinterest, Instagram, Google Plus, YouTube, AddThis, Disqus, GitHub, and more.</p>
                    """

    TRY_CASTLE = u"""
                    <h2 style="text-align: center;">TRY CASTLE CMS</h2>
                    <div class="row text-center">
                        <div class="col-xs-10 col-xs-offset-1 col-md-4 col-md-offset-1 col-lg-3 col-lg-offset-2">
                            <h4>Sign up for a free demo today, and see how Castle CMS can benefit your business.</h4>
                            <a data-val="410ec18805c44056813340350594a777" data-mce-href="resolveuid/410ec18805c44056813340350594a777" data-linktype="internal" href="https://castlecms.io/contact" class="btn btn-ghost" data-urltype="_direct_">Book a Demo</a>
                        </div>
                        <div class="col-md-2" style="margin: 20px 0;" data-mce-style="margin: 20px 0;">&nbsp;<img data-scale="mini" src="/++theme++castle.theme/img/frontpage/demo-icons.png" alt="Demo"></div><div class="col-xs-10 col-xs-offset-1 col-md-4 col-md-offset-0 col-lg-3">
                            <h4>Just looking for a quote? <br>Call <a href="tel:+17158693440" data-mce-href="tel:+17158693440">(715) 869-3440</a>, email <a href="mailto:info@wildcardcorp.com" data-mce-href="mailto:info@wildcardcorp.com">info@wildcardcorp.com</a>, or fill out <a data-val="410ec18805c44056813340350594a777" title="Demo and Quote Request" data-linktype="internal" href="https://castlecms.io/contact" data-mce-href="resolveuid/410ec18805c44056813340350594a777" data-urltype="_direct_">our form</a>.</h4>
                        </div>
                    </div>
                    """

    TWO_WIDE_1_1 = u"""
                    <h4>What technology is Castle CMS built on?</h4>
                    <p>The core of Castle CMS comes from <a title="Plone, the ultimate open source enterprise content management system" href="https://plone.com" target="_blank" data-linktype="external" data-urltype="/view" data-val="https://plone.com" data-mce-href="https://plone.com">Plone</a>, the open source CMS&nbsp;with <a title="Plone, the high security open source content management system" href="https://plone.com/secure" target="_blank" data-linktype="external" data-urltype="/view" data-val="https://plone.com/secure" data-mce-href="https://plone.com/secure">the best security track record</a>, bar none. Written in Python and using a dependable NoSQL database, Castle CMS is not vulnerable to the all-too-common attacks that afflict&nbsp;PHP and SQL systems. But Castle CMS is more than&nbsp;just that: it&nbsp;is a marriage of technologies and ready-to-go integrated services that provides a complete out of the box solution for organizations with&nbsp;sophisticated needs and demanding security and scalability&nbsp;requirements.&nbsp;</p>
                    """

    TWO_WIDE_1_2 = u"""
                    <h4>Who built Castle CMS?</h4>
                    <p>Castle CMS was developed&nbsp;by a highly experienced team&nbsp;from <a title="Wildcard Corp., the high security web technology solutions company" href="https://wildcardcorp.com" target="_blank" data-linktype="external" data-urltype="/view" data-val="https://wildcardcorp.com" data-mce-href="https://wildcardcorp.com">Wildcard Corp.</a>, a web development and hosting company located in Stevens Point, Wisconsin. Wildcard Corp. has a 10 year&nbsp;history of&nbsp;providing highly secure, robust systems to large government agencies. Our team includes core contributors and leaders&nbsp;of&nbsp;the open source <a title="Plone, the ultimate open source enterprise content management system" href="https://plone.com" target="_blank" data-linktype="external" data-urltype="/view" data-val="https://plone.com" data-mce-href="https://plone.com">Plone CMS</a>&nbsp;and <a title="The Plone community, bringing you the ultimate open source content management system" href="https://plone.org" target="_blank" data-linktype="external" data-urltype="/view" data-val="https://plone.org" data-mce-href="https://plone.org">community</a>. We combined our Plone and <a title="Wildcard Corp. success stories" href="https://wildcardcorp.com/success-stories" target="_blank" data-linktype="external" data-urltype="/view" data-val="https://wildcardcorp.com/success-stories" data-mce-href="https://wildcardcorp.com/success-stories">large government systems</a> experience to bring you&nbsp;an even better integrated CMS.</p>
                    """

    TWO_WIDE_2_1 = u"""
                    <h4>What kind of support is available for Castle CMS?</h4>
                    <p>Castle CMS is currently available as a tiered cloud-hosted solution, for corporate, local or state government, and federal government / enterprise levels. With the exception of specific integration components used by our largest clients, Castle CMS will be open sourced, giving you the freedom to inspect, enhance, and modify it, and the ability to avoid vendor lock-in. Wildcard Corp.&nbsp;has earned rave reviews from some of the most demanding organizations for its outstanding&nbsp;proactive support.&nbsp;For more information on our support levels, please fill out our no-obligation <a data-val="410ec18805c44056813340350594a777" title="Quote Request Form" data-linktype="internal" href="https://castlecms.io/contact" data-mce-href="resolveuid/410ec18805c44056813340350594a777" data-urltype="_direct_">quote request form</a>.</p>
                    """
    TWO_WIDE_2_2 = u"""
                    <h4>How much does Castle CMS cost?</h4>
                    <p>Each Castle CMS project is a custom-developed solution, and each client&#39;s needs vary greatly. We take great pride in being able to integrate Castle CMS with your existing systems and in providing you with design, customization, training, documentation, and support you can rely on.&nbsp;Please contact us&nbsp;for a no-obligation quote, either by emailing us at&nbsp;<a href="mailto:info@wildcardcorp.com?subject=Castle%20CMS%20Quote" data-linktype="email" data-urltype="/view" data-val="info@wildcardcorp.com" data-subject="Castle CMS Quote" data-mce-href="mailto:info@wildcardcorp.com?subject=Castle CMS Quote">info@wildcardcorp.com</a>,&nbsp;calling us at <a href="tel:+1%20(715)%20869-3440" data-linktype="external" data-urltype="/view" data-val="tel:+17158693440" data-mce-href="tel:+1 (715) 869-3440">+1 (715) 869-3440</a>, or filling out our <a data-val="410ec18805c44056813340350594a777" title="Quote Request Form" data-linktype="internal" href="https://castlecms.io/contact" data-mce-href="resolveuid/410ec18805c44056813340350594a777" data-urltype="_direct_">quote request form</a>. We look forward to hearing about&nbsp;your project requirements and being a part of your solution!</p>
                    """

    FAQS = u"""<h3 style="text-align: center;" data-mce-style="text-align: center;">FREQUENTLY ASKED QUESTIONS<br></h3>"""

    FEATURED_HIGHLIGHTS = u"""<h3 data-mce-style="text-align: center;" style="text-align: center;">FEATURE HIGHLIGHTS<br></h3>"""

    FRONTPAGE_TILES = [
        {
            'data': {
                'content': TOP_FULL_WIDTH
            },
            'meta': {
                'type': 'plone.app.standardtiles.rawhtml',
                'id': 'title-top'
            }
        },
        {
            'data': {
                'content': THREE_WIDE_1
            },
            'meta': {
                'type': 'plone.app.standardtiles.rawhtml',
                'id': 'three-wide-1'
            }
        },
        {
            'data': {
                'content': THREE_WIDE_2
            },
            'meta': {
                'type': 'plone.app.standardtiles.rawhtml',
                'id': 'three-wide-2'
            }
        },
        {
            'data': {
                'content': THREE_WIDE_3
            },
            'meta': {
                'type': 'plone.app.standardtiles.rawhtml',
                'id': 'three-wide-3'
            }
        },
        {
            'data': {
                'content': TRY_CASTLE
            },
            'meta': {
                'type': 'plone.app.standardtiles.rawhtml',
                'id': 'try-castle'
            }
        },
        {
            'data': {
                'content': TWO_WIDE_1_1
            },
            'meta': {
                'type': 'plone.app.standardtiles.rawhtml',
                'id': 'bottom-1'
            }
        },
        {
            'data': {
                'content': TWO_WIDE_1_2
            },
            'meta': {
                'type': 'plone.app.standardtiles.rawhtml',
                'id': 'bottom-2'
            }
        },
        {
            'data': {
                'content': TWO_WIDE_2_1
            },
            'meta': {
                'type': 'plone.app.standardtiles.rawhtml',
                'id': 'bottom-3'
            }
        },
        {
            'data': {
                'content': TWO_WIDE_2_2
            },
            'meta': {
                'type': 'plone.app.standardtiles.rawhtml',
                'id': 'bottom-4'
            }
        },
        {
            'data': {
                'content': FAQS
            },
            'meta': {
                'type': 'plone.app.standardtiles.rawhtml',
                'id': 'faqs'
            }
        },
        {
            'data': {
                'content': FEATURED_HIGHLIGHTS
            },
            'meta': {
                'type': 'plone.app.standardtiles.rawhtml',
                'id': 'featured-highlights'
            }
        },
        {
            'data': {
                'youtube_url': u'https://youtu.be/tSwQ6XkSvgU',
                'icon_link': u'/++theme++castle.theme/img/frontpage/icon-content-mgmt.png',
                'title': 'Intuitive Content Management',
                'text': RichTextValue("""<p>Creating, editing, and managing large amounts of content is easier than ever before with Castle's newly-designed toolbar. Use drag and drop to upload and organize content effortlessly on your website. Filter content, apply tags, set publication and expiry dates, review and publish items, and grant specific permissions to users and groups, all using our powerful content management view.</p>
                            <p>Need to move massive amounts of content around your site? Castle's asynchronous batch processing lets editors keep working as quickly as they need to.</p>
                            <p>Castle also supports any number of simultaneously logged-in users and editors with absolutely no additional licensing fees.</p>
                        """)
            },
            'meta': {
                'type': 'castle.cms.feature',
                'id': 'fp-feature-1'
            }
        },
        {
            'data': {
                'youtube_url': u'https://youtu.be/wjRNcsxSShQ',
                'icon_link': u'/++theme++castle.theme/img/frontpage/icon-security.png',
                'title': 'High Security',
                'text': RichTextValue("""<p>Castle CMS is based&nbsp;on Plone, which has&nbsp;the best security track record of any enterprise-scale content management system, but we have gone further, adding many&nbsp;security features to Castle:</p>
                            <ul>
                                <li>built-in two-factor authentication that protects you from&nbsp;stolen password exploits&nbsp;</li>
                                <li>auto-lockout: after a&nbsp;maximum number of login attempts, users are automatically locked out</li>
                                <li>user session management that allows administrators to terminate suspicious sessions</li>
                                <li>a login-secured, customizable dashboard&nbsp;for each user</li>
                                <li>integrated reCAPTCHA to keep spammers and bots&nbsp;at bay</li>
                                <li>personally identifiable information (PII) and metadata are automatically stripped from uploaded Office documents, PDFs, and other file types</li>
                            </ul>
                            <p>Castle&#39;s security is granular, letting&nbsp;you secure individual content items (pages, files, images, news items, calendar events) or entire sections of your site. If a user&nbsp;isn't&nbsp;authorized to access a content item, they won&#39;t even know it&#39;s there.</p>
                            <p>Castle's&nbsp;built-in search engine knows what each person is authorized to see, so you can rest easy knowing that private data will remain that way.</p>
                        """)
            },
            'meta': {
                'type': 'castle.cms.feature',
                'id': 'fp-feature-2'
            }
        },
        {
            'data': {
                'youtube_url': u'https://youtu.be/ZJR_9Zqup2k',
                'icon_link': u'/++theme++castle.theme/img/frontpage/icon-social-media.png',
                'title': 'Superior Social Media Integration',
                'text': RichTextValue("""<p>Embed tweets, Facebook statuses, timelines, and more as tiles that you can drag anywhere onto a page. Castle&nbsp;includes &ldquo;Share on Facebook&rdquo; and &ldquo;Share on Twitter&rdquo; buttons that you can place anywhere on&nbsp;a&nbsp;site. Castle&nbsp;even supports Twitter Cards, which allow you&mdash;and those that link to you&mdash;to attach rich photos, videos, and media to tweets. The best part about these features is that they are all available on any Castle site right out of the box.</p>
                            <p>Castle's built in search engine results are automatically adjusted&nbsp;according to&nbsp;your&nbsp;social media activity.</p>
                        """)
            },
            'meta': {
                'type': 'castle.cms.feature',
                'id': 'fp-feature-3'
            }
        },
        {
            'data': {
                'youtube_url': u'https://youtu.be/80bxHq-XmnY',
                'icon_link': u'/++theme++castle.theme/img/frontpage/icon-page-preview.png',
                'title': 'Content Previews',
                'text': RichTextValue("""<p>If one of your pages doesn&#39;t look right, you can find out before it&#39;s too late. Page Preview allows you to see what your published pages will look like while they&#39;re still private, which could save you lots of embarrassment.</p>
                        """)
            },
            'meta': {
                'type': 'castle.cms.feature',
                'id': 'fp-feature-4'
            }
        },
        {
            'data': {
                'youtube_url': u'https://youtu.be/w8KuF06T608',
                'icon_link': u'/++theme++castle.theme/img/frontpage/icon-layout-engine.png',
                'title': 'Dynamic Layout Engine',
                'text': RichTextValue("""<p>Create new layouts quickly and on a per&mdash;page basis. You can save and reuse&nbsp;them&nbsp;elsewhere on your site&nbsp;whenever you need to. Add columns as tiles that you can drag and drop into place.</p>
                        """)
            },
            'meta': {
                'type': 'castle.cms.feature',
                'id': 'fp-feature-5'
            }
        },
        {
            'data': {
                'youtube_url': u'https://youtu.be/Hj9qW34Hv4o',
                'icon_link': u'/++theme++castle.theme/img/frontpage/icon-easy-theming.png',
                'title': 'Easy Theming',
                'text': RichTextValue("""<p>Castle is extremely flexible for designers: its&nbsp;themes and designs are created&nbsp;using just&nbsp;a&nbsp;web browser, and you can&nbsp;set each individual page's design, or you can set a default design for sections of a site, or both. A Castle site can be themed to match any existing website and&nbsp;to fit into an integrated suite of web services, making it easy to integrate&nbsp;Castle-hosted content seamlessly.</p>
                        """)
            },
            'meta': {
                'type': 'castle.cms.feature',
                'id': 'fp-feature-6'
            }
        },
        {
            'data': {
                'youtube_url': u'https://youtu.be/zOVI65-TOI8',
                'icon_link': u'/++theme++castle.theme/img/frontpage/icon-adv-toolbar.png',
                'title': 'Intuitive Toolbar',
                'text': RichTextValue("""<p>Castle&#39;s toolbar was created in&mdash;house by our user experience team. It is&nbsp;designed to place all editing&nbsp;controls at your fingertips, intuitively and effectively.</p>
                        <p>Move content around, change its design or layout, share it with other users, access version history, examine its analytics data, even preview before you publish it.&nbsp;</p>""")
            },
            'meta': {
                'type': 'castle.cms.feature',
                'id': 'fp-feature-7'
            }
        },
        {
            'data': {
                'youtube_url': u'https://youtu.be/VakRVBIChOc',
                'icon_link': u'/++theme++castle.theme/img/frontpage/icon-search.png',
                'title': 'Comprehensive Built-in Search',
                'text': RichTextValue("""<p>Every item in your site&mdash;whether a page, Office document, PDF, news item, event, video, or audio&mdash;is automatically and instantly searchable from the built&mdash;in search bar. Integrated optical character recognition (OCR) for all PDF, Word, Excel,&nbsp;PowerPoint, and TIFF image&nbsp;files automatically makes all your uploaded documents searchable using&nbsp;the&nbsp;built in Elasticsearch engine, which is robust&nbsp;and&nbsp;fast&nbsp;enough to handle sites with millions of content items.</p>
                        <p>Tag your content with keywords&nbsp;to make everything even easier to find. You can also search for content according to its title, description, body text, author, editor, date, and more.</p>""")
            },
            'meta': {
                'type': 'castle.cms.feature',
                'id': 'fp-feature-8'
            }
        },
        {
            'data': {
                'youtube_url': u'https://youtu.be/SnOId_MT5H0',
                'icon_link': u'/++theme++castle.theme/img/frontpage/icon-seo.png',
                'title': 'Search Engine Optimization',
                'text': RichTextValue("""<p>Castle generates semantic&nbsp;HTML5, ensuring that Google and other search engines correctly interpret the structure and content of your&nbsp;website. You can&nbsp;achieve higher organic search results&nbsp;for any given page&nbsp;by&nbsp;setting&nbsp;SEO keywords and meta tags, which give hints to&nbsp;search engines as to what a web page is about.</p>
                        <p>Even if you have many contributors and editors, Castle helps you keep your site content under control by preventing anyone from uploading the same image or document twice&nbsp;and also by&nbsp;automatically uploading files&nbsp;to a central repository.</p>
                        <p>Your content can also be automatically archived after a set&nbsp;number of days, with&nbsp;each type of content&nbsp;having its own&nbsp;archival delay. Once archived, your content is stored in inexpensive Amazon Simple Storage Service (S3) bulk cloud storage, while&nbsp;remaining seamlessly available&nbsp;to your users at the same apparent web addresses. You can even have Castle move bulky content automatically to S3 if it exceeds a certain size.</p>""")
            },
            'meta': {
                'type': 'castle.cms.feature',
                'id': 'fp-feature-9'
            }
        },
        {
            'data': {
                'youtube_url': u'https://youtu.be/N3z9d76Wk1E',
                'icon_link': u'/++theme++castle.theme/img/frontpage/icon-doc-viewer.png',
                'title': 'A Viewer for All File Types',
                'text': RichTextValue("""<p>Castle's built-in&nbsp;document viewer works much like Adobe Reader, allowing&nbsp;you to view PDF, Word, Excel, and many other file types without ever leaving your browser. Your users won&#39;t need any locally installed programs to view or search through documents on your website.</p>
                        <p>Castle&#39;s document viewer includes dynamic search, so you can find a specific page within a document quickly and easily. It&nbsp;supports optical character recognition (OCR), turning scanned image files into searchable text documents. PDF conversion, available with the&nbsp;Castle Cloud hosted service, turns all uploaded Office files into platform&mdash;independent PDFs.</p>""")
            },
            'meta': {
                'type': 'castle.cms.feature',
                'id': 'fp-feature-10'
            }
        },
        {
            'data': {
                'youtube_url': u'https://youtu.be/zacn1repAFk',
                'icon_link': u'/++theme++castle.theme/img/frontpage/icon-quality-check.png',
                'title': 'Content Quality Assurance',
                'text': RichTextValue("""<p>Like your own personal proofreader, Castle can suggest content edits and provide inline content analysis checks. Castle includes&nbsp;a spell-check feature, a must&mdash;have&nbsp;for creating and maintaining high quality website content.</p>
                        <p>Castle's automatic content quality checking helps you&nbsp;improve&nbsp;content&nbsp;before it is published. This helps to improve your organic SEO.&nbsp;</p>
                        <p>Castle also provides content analytics out of the box like no other CMS. Examine analytics for the entire site, or just portions of your site,&nbsp;and see real&mdash;time statistics. Castle queries Google Analytics directly and puts the information at your fingertips by displaying&nbsp;it&nbsp;directly on the page you are editing.</p>""")
            },
            'meta': {
                'type': 'castle.cms.feature',
                'id': 'fp-feature-11'
            }
        },
        {
            'data': {
                'youtube_url': u'https://youtu.be/7U29U7KmlNI',
                'icon_link': u'/++theme++castle.theme/img/frontpage/icon-rich-content.png',
                'title': 'Rich Content Editing',
                'text': RichTextValue("""<p>Castle uses TinyMCE, the state of the art in what-you-see-is-what-you-get (WYSIWYG) web editors. Change formatting, bold text, italicize text, edit images, and even embed videos right from the editor. Your videos are automatically converted to a web compatible video format. You don&#39;t need to know CSS or HTML to&nbsp;write beautiful content. Everything is in plain English for your convenience.</p>
                        <p>But since we know not everyone&nbsp;reads&nbsp;English, Castle&#39;s editing and control interface supports more than 64 other languages as well, so international and multilingual users and content editors&nbsp;are at home&nbsp;on your&nbsp;website.</p>""")
            },
            'meta': {
                'type': 'castle.cms.feature',
                'id': 'fp-feature-12'
            }
        },
        {
            'data': {
                'fragment': 'coverimage'
            },
            'meta': {
                'type': 'castle.cms.fragment',
                'id': 'top-fragment'
            },
            'slot': 'top'
        }
    ]

    return FRONTPAGE_TILES



from AccessControl.SecurityManagement import newSecurityManager
from castle.cms._scripts import mjson
from castle.cms._scripts.importtypes import getImportType
from lxml.html import fromstring
from lxml.html import tostring
from OFS.interfaces import IFolder
from plone import api
from plone.app.blocks.layoutbehavior import ILayoutAware
from plone.app.redirector.interfaces import IRedirectionStorage
from plone.app.textfield.value import RichTextValue
from Products.CMFPlone.interfaces.constrains import ISelectableConstrainTypes
from zope.component import getUtility
from zope.component.hooks import setSite

import argparse
import logging
import os
import transaction


logger = logging.getLogger('castle.cms')

# ToDo:
#   - do NOT export images
#       - only images that are referenced in content
#   - lookup reference, if image only used once, add as lead image
#   - other images, if not referenced elsewhere, include inline
#       as data uris
#   - links that go to files should be moved to s3 and be converted
#       into file objects
#   - if default page and only page inside folder, replace folder
#


parser = argparse.ArgumentParser(
    description='...')
parser.add_argument('--site-id', dest='site_id', default='Plone')
parser.add_argument('--export-directory', dest='export_directory')
parser.add_argument('--overwrite', dest='overwrite', default=False)
args, _ = parser.parse_known_args()


user = app.acl_users.getUser('admin')  # noqa
newSecurityManager(None, user.__of__(app.acl_users))  # noqa
site = app[args.site_id]  # noqa
setSite(site)


def traverse(path):
    folder = site  # noqa
    for part in path.strip('/').split('/'):
        if part in folder.objectIds():
            folder = folder[part]
        else:
            return


def relpath(obj):
    return '/'.join(obj.getPhysicalPath())[len('/'.join(site.getPhysicalPath())) + 1:]


def recursive_create_path(path):
    if path in ('/', ''):
        folder = site
    else:
        path = path.lstrip('/')
        folder = traverse(path)
        if folder is not None and not IFolder.providedBy(folder):
            api.content.delete(folder)
            folder = None
        if folder is None:
            # Need to create folders up to where we want content
            # we'll walk it up create folders as needed
            folder = site
            for part in path.split('/'):
                try:
                    ob = folder[part]
                    if not IFolder.providedBy(ob):
                        api.content.delete(ob)
                        raise KeyError()
                    else:
                        folder = ob
                except (KeyError, AttributeError):
                    fpath = os.path.join(
                        args.export_directory,
                        relpath(folder),
                        part,
                        '__folder__')

                    if os.path.exists(fpath):
                        fi = open(fpath)
                        data = mjson.loads(fi.read())
                        fi.close()
                        importtype = getImportType(data, fpath[len(args.export_directory):], None)
                        creation_data = importtype.get_data()
                        creation_data['container'] = folder
                        creation_data['id'] = part
                        folder = api.content.create(**creation_data)
                        importtype.post_creation(folder)
                        if data['state']:
                            try:
                                api.content.transition(folder, to_state=data['state'])
                            except:
                                # maybe workflows do not match up
                                pass
                        folder.reindexObject()
                    else:
                        folder = api.content.create(
                            type='Folder',
                            id=part,
                            title=part.capitalize(),
                            container=folder)
                        bdata = ILayoutAware(folder)
                        bdata.contentLayout = '++contentlayout++castle/folder-query.html'
    return folder


_importable_fields = (
    'title',
    'description',
)


_read_phase = {}


def readObject(filepath):
    # disable, we don't use
    pass
    # fi = open(filepath)
    # data = mjson.loads(fi.read())
    # fi.close()


def fixHtmlImages(obj):
    try:
        html = obj.text.raw
    except:
        return
    if not html:
        return
    changes = False
    try:
        dom = fromstring(html)
    except:
        return
    for el in dom.cssselect('img'):
        src = el.attrib.get('src', '')
        if 'resolveuid' not in src:
            continue
        if '@@images' in src:
            continue
        parts = src.split('/')
        uid = parts[1]
        scale = '/'.join(parts[2:])
        if scale:
            scale = scale.replace('image_', '')
        src = 'resolveuid/%s/@@images/image' % uid
        if scale:
            src += '/' + scale
        attribs = {
            'src': src,
            'data-linktype': 'image',
            'data-val': uid
        }
        if scale:
            attribs['data-scale'] = scale
        el.attrib.update(attribs)
        changes = True
    if changes:
        obj.text = RichTextValue(
            tostring(dom), mimeType=obj.text.mimeType, outputMimeType=obj.text.outputMimeType)


def importObject(filepath, count):
    fi = open(filepath)
    data = mjson.loads(fi.read())
    fi.close()
    if filepath.endswith('__folder__'):
        filepath = '/'.join(filepath.split('/')[:-1])

    if data['portal_type'] in ('Topic', 'Collection', 'PressRoom'):
        return

    original_path = filepath[len(args.export_directory):]
    importtype = getImportType(data, original_path, _read_phase)
    path = importtype.get_path()
    folder = recursive_create_path('/'.join(path.split('/')[:-1]))
    _id = path.split('/')[-1]

    create = True
    if _id in folder.objectIds():
        if args.overwrite:
            if data['portal_type'] == 'Folder':
                cnt = folder[_id]
                if len(folder.objectIds()) == 1 or len(cnt.objectIds()) == 0:
                    api.content.delete(folder[_id])
            else:
                api.content.delete(folder[_id])
        else:
            create = False

    print('import path: %s' % path)
    creation_data = importtype.get_data()
    creation_data['container'] = folder

    aspect = ISelectableConstrainTypes(folder, None)
    if aspect:
        if (aspect.getConstrainTypesMode() != 1 or
                [creation_data['type']] != aspect.getImmediatelyAddableTypes()):
            aspect.setConstrainTypesMode(1)
            aspect.setImmediatelyAddableTypes([creation_data['type']])

    if create:
        obj = api.content.create(**creation_data)
    else:
        obj = folder[_id]
        for key, value in creation_data.items():
            if key not in ('id', 'type'):
                setattr(obj, key, value)

    if path != original_path:
        storage = getUtility(IRedirectionStorage)
        rpath = os.path.join('/'.join(site.getPhysicalPath()),
                             original_path.strip('/'))
        storage.add(rpath, "/".join(obj.getPhysicalPath()))

    importtype.post_creation(obj)
    if data['state']:
        try:
            api.content.transition(obj, to_state=data['state'])
        except:
            # maybe workflows do not match up
            pass

    fixHtmlImages(obj)
    obj.reindexObject()

    if count % 50 == 0:
        print('%i processed, committing' % count)
        transaction.commit()
        app._p_jar.invalidateCache()  # noqa
        transaction.begin()
        app._p_jar.sync()  # noqa


def importPages(path, count=0):
    for filename in os.listdir(path):
        if filename in ('.DS_Store', '__folder__'):
            continue
        count += 1
        filepath = os.path.join(path, filename)
        if os.path.isdir(filepath):
            count = importPages(filepath, count)
        else:
            try:
                importObject(filepath, count)
            except:
                logger.error('Error importing object', exc_info=True)
    return count


def importFolders(path, count=0):
    for filename in sorted(os.listdir(path)):
        count += 1
        filepath = os.path.join(path, filename)

        if filename == '__folder__':
            try:
                importObject(filepath, count)
            except:
                logger.error('Error importing object', exc_info=True)

        if os.path.isdir(filepath):
            count = importFolders(filepath, count)
    return count


def readPass(path):
    for filename in sorted(os.listdir(path)):
        if filename in ('.DS_Store', '__folder__'):
            continue
        filepath = os.path.join(path, filename)
        if os.path.isdir(filepath):
            readPass(filepath)
        else:
            readObject(filepath)


print('------------------------------')
print('Start importing')
print('------------------------------')
if args.overwrite:
    print('------------------------------')
    print('Importing with overwrite enabled')
    print('------------------------------')

print('doing read pass...')
# readPass(args.export_directory)
print('creating objects now...')
# importFolders(args.export_directory)
importPages(args.export_directory)
transaction.commit()



from zope.component import getAllUtilitiesRegisteredFor
from castle.cms.fragments.interfaces import IFragmentsDirectory


def getFragment(context, request, name):
    utils = getAllUtilitiesRegisteredFor(IFragmentsDirectory)
    utils.sort(key=lambda u: u.order)
    for util in reversed(utils):
        if util.layer is not None:
            if not util.layer.providedBy(request):
                continue
        try:
            return util.get(context, request, name)
        except KeyError:
            pass
    raise KeyError(name)



# -*- coding: utf-8 -*-
from zope.annotation.interfaces import IAnnotations
from plone.namedfile.file import NamedBlobImage
from castle.cms.testing import CASTLE_PLONE_INTEGRATION_TESTING
from castle.cms.tests.utils import get_tile
from castle.cms.tests.utils import render_tile
from plone import api
from plone.app.testing import login
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.uuid.interfaces import IUUID
from castle.cms.interfaces import IReferenceNamedImage
from castle.cms import lead
from plone.app.textfield.value import RichTextValue
from plone.namedfile.tests.test_image import zptlogo
from plone.tiles.data import ANNOTATIONS_KEY_PREFIX

import unittest


class TestLeadImages(unittest.TestCase):

    layer = CASTLE_PLONE_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer['portal']
        self.request = self.layer['request']
        login(self.portal, TEST_USER_NAME)
        setRoles(self.portal, TEST_USER_ID, ('Member', 'Manager'))
        self.folder_obj = api.content.create(
            type='Folder', container=self.portal, id='testfolder',)
        self.page_obj = api.content.create(
            type='Document', container=self.folder_obj, id='testpage',)
        self.image_obj = api.content.create(
            type='Image', container=self.folder_obj, id='testimage',
            image=NamedBlobImage(zptlogo, contentType='image/jpeg'))
        self.img_uid = IUUID(self.image_obj)

    def test_find_lead_image_in_text(self):
        self.page_obj.text = RichTextValue('''
<p>
<img src="resolveuid/{}/@@images/image/large" />
</p>
        '''.format(IUUID(self.image_obj)), mimeType='text/html', outputMimeType='text/html')
        image = lead.find_image(self.page_obj)
        self.assertEquals(IUUID(image), self.img_uid)

    def test_find_image_in_html(self):
        image = lead.find_image_in_html('''
<p>
<img src="resolveuid/{}/@@images/image/large" />
</p>
        '''.format(self.img_uid))
        self.assertEquals(IUUID(image), self.img_uid)

    def test_find_image_in_annotation(self):
        image = lead.find_image_in_annotation({
            'image': self.img_uid
        })
        self.assertEquals(IUUID(image), IUUID(self.image_obj))

    def test_find_image_in_annotation_on_object(self):
        annotations = IAnnotations(self.page_obj)
        annotations[ANNOTATIONS_KEY_PREFIX + 'foobar'] = {
            'image': self.img_uid
        }
        image = lead.find_image(self.page_obj)
        self.assertEquals(IUUID(image), self.img_uid)

    def test_find_image_in_annotation_with_html_on_object(self):
        annotations = IAnnotations(self.page_obj)
        annotations[ANNOTATIONS_KEY_PREFIX + 'foobar'] = {
            'content': '''
<p>
<img src="resolveuid/{}/@@images/image/large" />
</p>
            '''.format(self.img_uid)
        }
        image = lead.find_image(self.page_obj)
        self.assertEquals(IUUID(image), self.img_uid)

    def test_add_reference_image(self):
        self.page_obj.text = RichTextValue('''
<p>
<img src="resolveuid/{}/@@images/image/large" />
</p>
        '''.format(self.img_uid), mimeType='text/html', outputMimeType='text/html')
        lead.check_lead_image(self.page_obj)
        self.assertTrue(IReferenceNamedImage.providedBy(self.page_obj.image))
        self.assertEquals(self.page_obj.image.reference, self.img_uid)



from castle.cms import archival
from zope.component.hooks import setSite

import argparse
import json
import transaction


parser = argparse.ArgumentParser(
    description='...')
parser.add_argument('--file', dest='file', default=False)
parser.add_argument('--site-id', dest='site_id', default='Plone')
parser.add_argument('--site-url', dest='site_url', default='')
args, _ = parser.parse_known_args()

user = app.acl_users.getUser('admin')  # noqa
newSecurityManager(None, user.__of__(app.acl_users))  # noqa
site = app[args.site_id]  # noqa
setSite(site)


fi = open(args.file)
items = json.loads(fi.read())
fi.close()


storage = archival.Storage(site, UrlOpener=archival.RequestsUrlOpener)
count = 0
for item in items:
    count += 1
    content_path = '/' + '/'.join(item['path'].split('/')[2:])
    url = args.site_url.rstrip('/') + content_path
    new_url = storage.add_url(url, content_path, item['uid'])  # need to export UID also
    if new_url:
        print('imported %s -> %s' % (url, new_url))
    else:
        print('error importing %s' % (url,))
    if count % 100 == 0:
        print('done with %i' % count)
        transaction.commit()

transaction.commit()



from AccessControl import getSecurityManager
from Acquisition import aq_parent
from castle.cms.interfaces import ITrashed
from plone import api
from plone.app.dexterity.behaviors.nextprevious import NextPreviousBase
from plone.registry.interfaces import IRegistry
from Products.CMFCore.interfaces._content import IFolderish
from Products.CMFPlone.interfaces.siteroot import IPloneSiteRoot
from zope.component import getUtility


class NextPrevious(NextPreviousBase):

    def __init__(self, context):
        if not IFolderish.providedBy(context):
            # need to be able to drop on page as well
            context = aq_parent(context)

        self.context = context
        registry = getUtility(IRegistry)
        self.vat = registry.get('plone.types_use_view_action_in_listings', [])
        self.security = getSecurityManager()
        if IPloneSiteRoot.providedBy(context):
            catalog = api.portal.get_tool('portal_catalog')
            folder_path = '/'.join(context.getPhysicalPath())
            brains = catalog(path={'query': folder_path, 'depth': 1})
            order = [b.id for b in brains]
        else:
            order = context.getOrdering()
            if not isinstance(order, list):
                order = order.idsInOrder()
            if not isinstance(order, list):
                order = None
        self.order = order

    def getNextItem(self, obj):
        """ return info about the next item in the container """
        if not self.order:
            return None
        try:
            pos = self.order.index(obj.getId())
        except ValueError:
            return
        for oid in self.order[pos + 1:]:
            data = self.getData(self.context[oid])
            if data:
                return data

    def getData(self, obj):
        if ITrashed.providedBy(obj):
            return None
        data = super(NextPrevious, self).getData(obj)
        if data:
            data['obj'] = obj
        return data



# -*- coding: utf-8 -*-
from lxml.html import fromstring
import unittest

from DateTime import DateTime
from castle.cms import archival
from castle.cms.testing import CASTLE_PLONE_INTEGRATION_TESTING
from plone import api
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.app.testing import login
from plone.app.testing import setRoles


class TestArchival(unittest.TestCase):

    layer = CASTLE_PLONE_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer['portal']
        self.request = self.layer['request']

    def test_get_archival_items(self):
        login(self.portal, TEST_USER_NAME)
        setRoles(self.portal, TEST_USER_ID, ('Member', 'Manager'))
        foobar = api.content.create(
            type='Document',
            id='foobar1',
            container=self.portal)
        api.content.transition(foobar, 'publish')
        foobar.setModificationDate(DateTime() - 10)
        foobar.reindexObject(idxs=['modified'])
        api.portal.set_registry_record(
            'castle.archival_number_of_days', 5)
        api.portal.set_registry_record(
            'castle.archival_types_to_archive', ['Document'])

        self.assertEqual(len(archival.getContentToArchive()), 1)

    def test_get_archival_items_pays_attention_to_types(self):
        login(self.portal, TEST_USER_NAME)
        setRoles(self.portal, TEST_USER_ID, ('Member', 'Manager'))
        foobar = api.content.create(
            type='Document',
            id='foobar1',
            container=self.portal)
        api.content.transition(foobar, 'publish')
        foobar.setModificationDate(DateTime() - 10)
        foobar.reindexObject(idxs=['modified'])
        api.portal.set_registry_record(
            'castle.archival_number_of_days', 5)
        api.portal.set_registry_record(
            'castle.archival_types_to_archive', ['News Item'])

        self.assertEqual(len(archival.getContentToArchive()), 0)

    def test_flash_resource_mover_gets_els(self):
        dom = fromstring('''<html>
<head>
<base href="https://www.foobar.com/foobar/"/><!--[if lt IE 7]></base><![endif]-->
</head><body>
<script type="text/javascript">
    AC_FL_RunContent('codebase','http://download.macromedia.com/pub/shockwave/cabs/flash/swflash.cab#version=9,0,28,0','width','920','height','670','src','foobar','quality','high','wmode','transparent','pluginspage','http://www.adobe.com/shockwave/download/download.cgi?P1_Prod_Version=ShockwaveFlash','movie','foobar' );
    </script></body></html>''')
        mover = archival.FlashScriptResourceMover(dom)
        els = mover.get_elements()
        self.assertEqual(len(els), 1)
        self.assertEqual(mover.get_url(els[0]),
                         'https://www.foobar.com/foobar/foobar.swf')
        mover.modify(els[0], "https://www.foobar.com/foobar/foobar.swf")
        self.assertTrue("https://www.foobar.com/foobar/foobar'" in els[0].text)

    def test_flash_resource_mover_gets_els_bad_markup(self):
        dom = fromstring('''<html>
<head>
<base href="https://www.foobar.com/foobar/avatars/avatars"/><!--[if lt IE 7]></base><![endif]-->
</head><body>
<script type="text/javascript">
    // &amp;amp;amp;amp;amp;amp;amp;amp;amp;amp;amp;lt;![CDATA[
    AC_FL_RunContent( 'codebase','http://download.macromedia.com/pub/shockwave/cabs/flash/swflash.cab#version=9,0,28,0','width','630','height','341','src','avatars_2','quality','high', 'wmode','transparent','pluginspage','http://www.adobe.com/shockwave/download/download.cgi?P1_Prod_Version=ShockwaveFlash','movie','avatars_2?12' ); //end AC code
    // ]]&amp;amp;amp;amp;amp;amp;amp;amp;amp;amp;amp;gt;
    </script></body></html>''')
        mover = archival.FlashScriptResourceMover(dom)
        els = mover.get_elements()
        self.assertEqual(len(els), 1)
        self.assertEqual(mover.get_url(els[0]),
                         'https://www.foobar.com/foobar/avatars/avatars_2.swf')
        mover.modify(els[0], "https://www.foobar.com/foobar/avatars/avatars_2.swf")
        self.assertTrue("https://www.foobar.com/foobar/avatars/avatars_2'" in els[0].text)

    def test_archive_replacement_text(self):
        login(self.portal, TEST_USER_NAME)
        setRoles(self.portal, TEST_USER_ID, ('Member', 'Manager'))
        storage = archival.Storage(self.portal)

        storage.replacements = {
            'foobar': 'barfoo'
        }

        self.assertEqual('<html><body><div>barfoo</div></body></html>',
                         storage.apply_replacements('<html><body><div>foobar</div>'
                                                    '</body></html>'))

    def test_archive_replacement_selector(self):
        login(self.portal, TEST_USER_NAME)
        setRoles(self.portal, TEST_USER_ID, ('Member', 'Manager'))
        storage = archival.Storage(self.portal)

        storage.replacements = {
            '.foobar': 'barfoo'
        }

        self.assertEqual('<html><body><div class="foobar">barfoo</div></body></html>',
                         storage.apply_replacements('<html><body><div class="foobar"></div>'
                                                    '</body></html>'))

    def test_archive_transformers(self):
        login(self.portal, TEST_USER_NAME)
        setRoles(self.portal, TEST_USER_ID, ('Member', 'Manager'))
        storage = archival.Storage(self.portal)
        result = storage.transform_content(
            '<html><body><div class="foo">foo</div></body></html>',
            'http://foobar.com')

        self.assertTrue('>bar<' in result)



from AccessControl.SecurityManagement import newSecurityManager
from castle.cms.utils import retriable
from datetime import datetime
from plone import api
from plone.registry.interfaces import IRegistry
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.interfaces.siteroot import IPloneSiteRoot
from tendo import singleton
from zope.component import getUtility
from zope.component.hooks import getSite
from zope.component.hooks import setSite

import logging
import transaction


logger = logging.getLogger('castle.cms')


@retriable(sync=True)
def delete_user(user):
    # need to clean/change ownership of content
    admin_user = api.user.get('admin')
    if admin_user:
        admin_user = admin_user.getUser()
        catalog = api.portal.get_tool('portal_catalog')
        for brain in catalog(Creator=user.getId()):
            ob = brain.getObject()
            ob.changeOwnership(admin_user, recursive=False)
            ob.reindexObjectSecurity()

    site = getSite()
    acl_users = site.acl_users
    if not user.canDelete():
        return
    try:
        acl_users.user_folderDelUsers([user.getId()])
    except (AttributeError, NotImplementedError):
        pass

    # Delete member data in portal_memberdata.
    mdtool = getToolByName(site, 'portal_memberdata', None)
    mtool = getToolByName(site, 'portal_membership')
    if mdtool is not None:
        mdtool.deleteMemberData(user.getId())

    # Delete members' local roles.
    mtool.deleteLocalRoles(site, [user.getId()], reindex=1, recursive=1)


def clean(site):
    setSite(site)
    registry = getUtility(IRegistry)
    days_to_clean_disabled_users = registry.get(
        'plone.days_to_clean_disabled_users', 30)
    for user in api.user.get_users():
        if api.user.get_roles(user=user) == ['Authenticated']:
            # compare when user was last modified
            diff = datetime.utcnow() - datetime.fromtimestamp(user._p_mtime)
            minutes = diff.seconds / 60
            hours = minutes / 60
            days = hours / 24
            if days >= days_to_clean_disabled_users:
                delete_user(user)
    transaction.commit()


def run(app):
    singleton.SingleInstance('cleanusers')

    user = app.acl_users.getUser('admin')  # noqa
    newSecurityManager(None, user.__of__(app.acl_users))  # noqa

    for oid in app.objectIds():  # noqa
        obj = app[oid]  # noqa
        if IPloneSiteRoot.providedBy(obj):
            try:
                clean(obj)
            except:
                logger.error('Could not clean users %s' % oid, exc_info=True)


if __name__ == '__main__':
    run(app)  # noqa



from BTrees.OOBTree import OOBTree
from plone import api
from plone.app.uuid.utils import uuidToObject
from plone.uuid.interfaces import IUUID
from zope.annotation.interfaces import IAnnotations


DATA_KEY = 'castle.cms.filehashes'


class DuplicateDetector(object):

    def __init__(self):
        self.site = api.portal.get()
        self.annotations = IAnnotations(self.site)
        self.data = self.annotations.get(DATA_KEY)

    def register(self, obj, hash):
        if self.data is None:
            self.annotations[DATA_KEY] = OOBTree()
            self.data = self.annotations[DATA_KEY]
        self.data[hash] = IUUID(obj)

    def get_object(self, hash):
        if self.data is None:
            return None
        if hash not in self.data:
            return None
        uuid = self.annotations[DATA_KEY][hash]
        return uuidToObject(uuid)


class DuplicateException(Exception):

    def __init__(self, obj):
        self.obj = self.object = self.context = obj
        self.message = 'File already found'



from castle.cms import constants
from castle.cms.vocabularies import BusinessTypesVocabulary
from plone.keyring import django_random
from Products.CMFPlone.interfaces import controlpanel
from Products.CMFPlone.utils import validate_json
from zope import schema
from zope.interface import Interface
from zope.schema.vocabulary import SimpleVocabulary

def create_term(val, label):
    return SimpleVocabulary.createTerm(val, val, label)


class ISocialMediaSchema(controlpanel.ISocialMediaSchema):
    twitter_consumer_key = schema.TextLine(
        title=u'Twitter consumer key',
        default=None,
        required=False)

    twitter_consumer_secret = schema.TextLine(
        title=u'Twitter consumer secret',
        default=None,
        required=False)

    twitter_oauth_token = schema.TextLine(
        title=u'Twitter OAuth token',
        required=False)

    twitter_oauth_secret = schema.TextLine(
        title=u'Twitter OAuth secret',
        required=False)

    twitter_screen_name = schema.TextLine(
        title=u'Twitter Screen Name',
        required=False)

    twitter_timeline_widget = schema.TextLine(
        title=u'Twitter widget ID',
        required=True,
        default=u'684100313582833665')

    disquis_shortname = schema.TextLine(
        title=u'Disqus Shortname',
        description=u'If provided, Disqus will be used instead of '
                    u"Plone's default commenting mechanism. You need "
                    u'to sign up for disqus in order for this feature to work',
        required=False,
        default=None)

    youtube_channel = schema.URI(
        title=u'Youtube channel URL',
        required=False
    )

    google_plus_username = schema.TextLine(
        title=u'Google+ username',
        required=False
    )

    github_username = schema.TextLine(
        title=u'Github Username',
        required=False
    )


class ISiteSchema(controlpanel.ISiteSchema):
    short_site_title = schema.TextLine(
        title=u'Short site title',
        default=u"Castle",
        required=False)

    site_icon = schema.ASCII(
        title=u"Site Icon",
        description=u'Castle will use this icon to generate all the various '
                    u'icons for mobile devices necessary',
        required=False,
    )

    public_url = schema.TextLine(
        title=u'Public site url',
        default=None,
        required=False)

    backend_url = schema.TextLine(
        title=u'Backend site url',
        default=u'',
        required=False)

    system_email_addresses = schema.List(
        title=u"System email addresses",
        description=u'List of admin email addresses',
        default=[],
        missing_value=[],
        value_type=schema.TextLine(),
        required=False)

    disable_contact_form = schema.Bool(
        title=u'Disable contact form',
        default=False)

    enable_notification_subscriptions = schema.Bool(
        title=u'Enable notification subscriptions',
        description=u'By going to @@subscribe',
        default=False)


class ISecuritySchema(controlpanel.ISecuritySchema):
    failed_login_attempt_window = schema.Int(
        title=u'Failed login attempts window(seconds)',
        required=True,
        default=900)

    max_failed_login_attempts = schema.Int(
        title=u'Max failed login attempts',
        required=True,
        default=8)

    days_to_clean_disabled_users = schema.Int(
        title=u'Number of days before disabled users are removed',
        default=30
    )

    login_shield_setting = schema.Choice(
        title=u'Login shield',
        description=u'Require user to login before they can even '
                    u'access the site',
        default=constants.SHIELD.NONE,
        vocabulary=SimpleVocabulary([
            create_term(constants.SHIELD.NONE, 'Not active'),
            create_term(constants.SHIELD.BACKEND, 'Protect only backend urls'),
            create_term(constants.SHIELD.ALL, 'Protect all traffic. BE CAREFUL ACTIVATING THIS.'),
        ])
    )

    two_factor_enabled = schema.Bool(
        title=u'Require two factor authentication',
        description=u'Requires users to have email and/or phone number set',
        default=False,
    )

    restrict_logins_to_countries = schema.Tuple(
        title=u'Restrict logins to countries',
        description=u'Choose countries that logins should be restricted to. '
                    u'This feature only works if your proxy server is '
                    u'forwarding the user country code to the Castle server. '
                    u'Leave empty to allow all logins.',
        missing_value=(),
        default=(),
        required=False,
        value_type=schema.Choice(
            vocabulary='castle.cms.vocabularies.Countries'
        )
    )


class IAnnoucementData(Interface):
    show_announcement = schema.Bool(
        title=u'Show site announcement',
        default=False,
        required=False)

    site_announcement = schema.Text(
        title=u"Site announcement",
        default=u'<p><strong>Breaking News:</strong> '
                u'<a href="#" style="color: white;">Follow Updates from...</a></p>',
        required=False
    )

    subscriber_categories = schema.List(
        title=u'Subscription Categories',
        description=u"Categories that users can subscribe to. One per line",
        default=[],
        missing_value=[],
        required=False,
        value_type=schema.TextLine())


class ISiteConfiguration(Interface):

    image_repo_location = schema.TextLine(
        title=u'Image repo path',
        required=True,
        default=u'/image-repository')

    file_repo_location = schema.TextLine(
        title=u'File repo path',
        required=True,
        default=u'/file-repository')

    audio_repo_location = schema.TextLine(
        title=u'Audio repo location',
        required=True,
        default=u'/audio-repository')

    video_repo_location = schema.TextLine(
        title=u'Video repo location',
        required=True,
        default=u'/video-repository')

    max_file_size = schema.Int(
        title=u'Max file size(MB)',
        description=u'Before uploading to S3 storage',
        default=50)

    allowed_locations = schema.List(
        title=u'Allowed locations',
        default=[],
        missing_value=[],
        required=False,
        value_type=schema.TextLine())

    publish_require_quality_check = schema.Bool(
        title=u'Require quality check in publishing',
        description=u'Require quality check passing before content can be published. If it '
                    u'does not pass, user needs to add comment in order to override.',
        default=False
    )


class IAPISettings(Interface):
    princexml_server_url = schema.TextLine(
        title=u'PrinceXML server url',
        description=u'required in order to convert documents',
        default=u'http://localhost:6543/convert',
        required=False)

    google_maps_api_key = schema.TextLine(
        title=u'Google Maps API Key',
        default=None,
        required=False)

    google_api_email = schema.TextLine(
        title=u'Google API Email',
        default=None,
        required=False)

    google_api_service_key_file = schema.ASCII(
        title=u"Google API Service Key File",
        description=u'Private key file',
        required=False,
    )

    google_analytics_id = schema.TextLine(
        title=u'Google Analytics ID',
        description=u'for use with gathering content statistics',
        required=False)

    recaptcha_public_key = schema.TextLine(
        title=u'Recaptcha 3 Public Key',
        required=False)

    recaptcha_private_key = schema.TextLine(
        title=u'Recaptcha 3 Private Key',
        required=False)

    aws_s3_key = schema.TextLine(
        title=u'AWS S3 Key',
        required=False,
        default=None)

    aws_s3_secret = schema.TextLine(
        title=u'AWS S3 Secret',
        required=False,
        default=None)

    aws_s3_bucket_name = schema.TextLine(
        title=u'AWS S3 Bucket',
        required=False,
        default=None)

    aws_s3_host_endpoint = schema.TextLine(
        title=u'AWS Host endpoint',
        description=u'Leave empty unless you know what you are doing here.',
        required=False,
        default=None)

    aws_s3_base_url = schema.TextLine(
        title=u'AWS File Base Url',
        description=u'If you are providing your own domain name to serve files from',
        required=False,
        default=None)

    plivo_auth_id = schema.TextLine(
        title=u'Plivo Auth ID',
        description=u'Text messaging API',
        required=False)

    plivo_auth_token = schema.TextLine(
        title=u'Plivo Auth Token',
        required=False)

    plivo_phone_number = schema.TextLine(
        title=u'Plivo Source Number',
        description=u'For making text messages from',
        required=False)

    etherpad_url = schema.TextLine(
        title=u'Etherpad Url',
        required=False
    )

    etherpad_api_key = schema.TextLine(
        title=u'Etherpad API Key',
        required=False
    )

    cf_api_key = schema.TextLine(
        title=u'Cloudflare API Key',
        description=u'Setting an API Key here and enabling cache purging '
                    u'activates purging against Cloudflare.',
        required=False
    )

    cf_email = schema.TextLine(
        title=u'Cloudflare Email',
        description=u'One associated with cloudflare api key',
        required=False
    )

    cf_zone_id = schema.TextLine(
        title=u'Cloudflare Zone ID',
        required=False)

    rocket_chat_front_page = schema.TextLine(
        title=u'Rocket.Chat User URL',
        description=u'URL of the Rocket.Chat server to connect to',
        required=False
    )

    rocket_chat_secret = schema.TextLine(
        title=u'Rocket.Chat secret',
        description=u'Text string used to salt Rocket.Chat authentication tokens',
        required=False,
        default=unicode(django_random.get_random_string(64))
    )

class IArchivalSettings(Interface):
    archival_enabled = schema.Bool(
        title=u'Archival Enabled',
        description=u'Archives are stored in configured AWS bucket',
        default=False)

    archival_number_of_days = schema.Int(
        title=u'Archival number of days from last edit',
        description=u'For the archival to find content to archive',
        default=365 * 3)

    archival_types_to_archive = schema.List(
        title=u'Types to archive',
        default=[
            'News Item'
        ],
        missing_value=[],
        value_type=schema.Choice(
            vocabulary="castle.cms.vocabularies.ReallyUserFriendlyTypes"))

    archival_states_to_archive = schema.List(
        title=u'States to archive',
        description=u'These MUST be states that allow anonymous users to view the content. '
                    u'Any states that do not, will be ignored',
        default=[
            'published'
        ],
        missing_value=[],
        value_type=schema.Choice(
            vocabulary="plone.app.vocabularies.WorkflowStates"))

    archival_replacements = schema.Dict(
        title=u'Archival replacements',
        description=u'Replacement text that will be applied to the content being archived.',
        default={},
        missing_value={},
        required=False,
        key_type=schema.TextLine(),
        value_type=schema.TextLine())


class ICastleSettings(
        ISiteConfiguration,
        IAPISettings,
        IAnnoucementData,
        IArchivalSettings):
    pass


class IBusinessData(Interface):
    business_type = schema.Choice(
        title=u'Business Type',
        vocabulary=BusinessTypesVocabulary,
        default=None,
        required=False
    )

    business_name = schema.TextLine(
        title=u'Business Name',
        required=False)

    business_address = schema.TextLine(
        title=u'Address',
        required=False)

    business_city = schema.TextLine(
        title=u'City',
        required=False)

    business_state = schema.TextLine(
        title=u'State',
        required=False)

    business_postal_code = schema.TextLine(
        title=u'Postal Code',
        required=False)

    business_country = schema.TextLine(
        title=u'Country',
        required=False,
        default=u'US')

    business_coordinates = schema.Text(
        title=u'Coordinates',
        description=u'Search for address to get the coordinates of business address',
        default=u'{}',
        required=False,
        constraint=validate_json
    )

    business_telephone = schema.TextLine(
        title=u'Telephone',
        required=False)

    business_days_of_week = schema.List(
        title=u'Days open',
        required=False,
        default=[],
        value_type=schema.Choice(
            vocabulary=SimpleVocabulary([
                SimpleVocabulary.createTerm('Monday', 'Monday', 'Monday'),
                SimpleVocabulary.createTerm('Tuesday', 'Tuesday', 'Tuesday'),
                SimpleVocabulary.createTerm('Wednesday', 'Wednesday', 'Wednesday'),
                SimpleVocabulary.createTerm('Thursday', 'Thursday', 'Thursday'),
                SimpleVocabulary.createTerm('Friday', 'Friday', 'Friday'),
                SimpleVocabulary.createTerm('Saturday', 'Saturday', 'Saturday'),
                SimpleVocabulary.createTerm('Sunday', 'Sunday', 'Sunday')
            ])
        )
    )

    business_opens = schema.TextLine(
        title=u'Opens',
        description=u'In the format of 08:00',
        required=False)

    business_closes = schema.TextLine(
        title=u'Closes',
        description=u'In the format of 08:00',
        required=False)

    business_special_hours = schema.List(
        title=u'Special hours',
        description=u'In the format Day|Opens|Closes. Example: Saturday|08:00|12:00',
        value_type=schema.TextLine(),
        default=[],
        required=False)

    business_menu_link = schema.TextLine(
        title=u'Menu link',
        description=u'If your business has a menu, provide link here',
        required=False)

    business_accepts_reservations = schema.Bool(
        title=u'Accepts Reservations',
        default=False)

    business_additional_configuration = schema.Text(
        title=u'Additional Configuration',
        description=u'See https://developers.google.com/structured-data/local-businesses/ '
                    u'for details on how to configuration business search data',
        constraint=validate_json,
        required=False)


class ICrawlerConfiguration(Interface):
    crawler_active = schema.Bool(
        title=u'Active',
        required=False,
        default=False)

    crawler_index_archive = schema.Bool(
        title=u'Index archives',
        required=False,
        default=False
    )

    crawler_site_maps = schema.List(
        title=u'Site maps',
        required=False,
        value_type=schema.TextLine()
    )

    crawler_user_agent = schema.TextLine(
        title=u'User Agent',
        description=u'User agent to use when crawling sites',
        default=u'Castle CMS Crawler 1.0'
    )



from castle.cms import archival
from castle.cms.utils import retriable
from collective.celery import task
from plone.uuid.interfaces import IUUID
from zope.component.hooks import getSite


def _sync_and_store(storage, uid):
    # need to be a bit tricky here since this could take a while to do,
    # we want to re-sync db and THEN commit...
    data = storage.archives[uid]
    storage.site._p_jar.sync()

    storage.archives[uid] = data
    storage.path_to_uid[data['path']] = uid


@retriable()
def _archive_url(url, path, uid):
    storage = archival.Storage(getSite(), UrlOpener=archival.RequestsUrlOpener)
    storage.add_url(url, path, uid)
    _sync_and_store(storage, uid)


@retriable()
def _archive_content(obj):
    storage = archival.Storage(getSite())
    storage.add_content(obj)
    _sync_and_store(storage, IUUID(obj))


@task.as_admin()
def archive_url(url, path, uid):
    return _archive_url(url, path, uid)


@task.as_admin()
def archive_content(obj):
    return _archive_content(obj)



from Acquisition import aq_parent
from castle.cms import cache
from castle.cms.interfaces import IAuthenticator
from castle.cms.interfaces import ICastleApplication
from castle.cms.lockout import LockoutManager
from castle.cms.utils import get_ip
from castle.cms.utils import get_random_string
from castle.cms.utils import strings_differ
from OFS.interfaces import IItem
from plone import api
from plone.keyring.interfaces import IKeyManager
from plone.keyring.keymanager import KeyManager
from plone.registry.interfaces import IRegistry
from plone.session.plugins.session import manage_addSessionPlugin
from Products.PlonePAS.events import UserLoggedInEvent
from Products.PlonePAS.setuphandlers import activatePluginInterfaces
from Products.PlonePAS.setuphandlers import migrate_root_uf
from Products.PluggableAuthService.interfaces.plugins import IChallengePlugin
from zope.component import adapter
from zope.component import getGlobalSiteManager
from zope.component import getUtility
from zope.component.interfaces import ComponentLookupError
from zope.event import notify
from zope.interface import implementer
from zope.publisher.interfaces.browser import IDefaultBrowserLayer

import time


class AuthenticationException(Exception):
    pass


class AuthenticationCountryBlocked(AuthenticationException):
    pass


class AuthenticationMaxedLoginAttempts(AuthenticationException):
    pass


class AuthenticationUserDisabled(AuthenticationException):
    pass


class AuthenticationPasswordResetWindowExpired(AuthenticationException):
    pass


@implementer(IAuthenticator)
@adapter(IItem, IDefaultBrowserLayer)
class Authenticator(object):

    def __init__(self, context, request):
        self.context = context
        self.request = request
        try:
            self.registry = getUtility(IRegistry)
            self.two_factor_enabled = self.registry.get(
                'plone.two_factor_enabled', False)
        except ComponentLookupError:
            self.registry = None
            self.two_factor_enabled = False

    @property
    def is_zope_root(self):
        return ICastleApplication.providedBy(self.context)

    def get_tool(self, name):
        if self.is_zope_root:
            return getattr(self.context, name, None)
        else:
            return api.portal.get_tool(name)

    def get_supported_auth_schemes(self):
        auth_schemes = [{
            'id': 'email',
            'label': 'Email'
        }]
        if self.registry:
            if self.registry.get('castle.plivo_auth_id'):
                auth_schemes.append({
                    'id': 'sms',
                    'label': 'SMS'
                })
        return auth_schemes

    def authorize_2factor(self, username, code, offset=0):
        try:
            value = cache.get(self.get_2factor_code_key(username))
        except:
            return False

        # check actual code
        if strings_differ(value['code'].lower(), code.lower()):
            return False

        # then check timing
        timestamp = value.get('timestamp')
        if not timestamp or (time.time() > (timestamp + 5 * 60 + offset)):
            return False
        return True

    def get_2factor_code_key(self, username):
        return '{}-{}-auth-code-attempt'.format(
            '-'.join(self.context.getPhysicalPath()[1:]),
            username
        )

    def issue_2factor_code(self, username):
        cache_key = self.get_2factor_code_key(username)
        code = get_random_string(8).upper()
        # store code to check again later
        cache.set(cache_key, {
            'code': code,
            'timestamp': time.time()
        })
        return code

    def authenticate(self, username=None, password=None, country=None):
        """
        return true if successfull
        """
        if not self.is_zope_root:
            manager = LockoutManager(self.context, username)

            if manager.maxed_number_of_attempts():
                raise AuthenticationMaxedLoginAttempts()

            manager.add_attempt()

        for acl_users in self.get_acl_users():
            # if not root, could be more than one to check against
            user = acl_users.authenticate(username, password, self.request)
            if user:
                break

        if user is None:
            return False, user

        if not self.is_zope_root:
            manager.clear()

        if user.getRoles() == ['Authenticated']:
            raise AuthenticationUserDisabled()

        if self.registry:
            allowed_countries = self.registry.get(
                'plone.restrict_logins_to_countries')
            if allowed_countries and country:
                if country not in allowed_countries:
                    if not self.country_exception_granted(user.getId()):
                        raise AuthenticationCountryBlocked()

        if not self.is_zope_root:
            member = api.user.get(user.getId())
            reset_password = member.getProperty(
                'reset_password_required', False)
            reset_time = member.getProperty('reset_password_time', None)

            if reset_password and reset_time:
                if reset_time + (24 * 60 * 60) < time.time():
                    raise AuthenticationPasswordResetWindowExpired()

        acl_users.session._setupSession(user.getId(), self.request.response)
        notify(UserLoggedInEvent(user))

        return True, user

    def country_exception_granted(self, userid):
        cache_key = self.get_country_exception_cache_key(userid)
        try:
            data = cache.get(cache_key)
        except:
            return False
        timestamp = data.get('timestamp')
        if (data.get('granted') and timestamp and
                (time.time() < (timestamp + (12 * 60 * 60)))):
            return True
        return True

    def get_acl_users(self):
        """
        get list of acl_user objects,
        first, site, then root
        """
        objects = [self.get_tool('acl_users')]
        if not self.is_zope_root:
            context = aq_parent(self.context)
            while context and not ICastleApplication.providedBy(context):
                context = aq_parent(context)
            acl = getattr(context, 'acl_users', None)
            if acl:
                objects.append(acl)
        return objects

    def get_country_exception_cache_key(self, userid):
        return '{}-{}-country-exception'.format(
            '-'.join(self.context.getPhysicalPath()[1:]),
            userid
        )

    def issue_country_exception(self, user, country):
        # capture information about the request
        data = {
            'referrer': self.request.get_header('REFERER'),
            'user_agent': self.request.get_header('USER_AGENT'),
            'ip': get_ip(self.request),
            'username': user.getUserName(),
            'userid': user.getId(),
            'country': country,
            'timestamp': time.time(),
            'code': get_random_string(50),
            'granted': False
        }

        cache_key = self.get_country_exception_cache_key(user.getId())
        cache.set(cache_key, data, 12 * 60 * 60)  # valid for 12 hours

        return data

    def change_password(self, member, new_password):
        user = api.user.get(member.getId())
        user.setMemberProperties(mapping={
            'reset_password_required': False,
            'reset_password_time': time.time()
        })
        member.setSecurityProfile(password=new_password)


def install_acl_users(app, logger=None):
    uf = app.acl_users
    found = uf.objectIds(['Plone Session Plugin'])
    if not found:
        # new root acl user implementation not installed yet
        migrate_root_uf(app)
        uf = app.acl_users  # need to get new acl_users

        plone_pas = uf.manage_addProduct['PlonePAS']
        manage_addSessionPlugin(plone_pas, 'session')
        activatePluginInterfaces(app, "session")

        cookie_auth = uf.credentials_cookie_auth
        cookie_auth.login_path = u'/@@secure-login'

        uf.plugins.activatePlugin(
            IChallengePlugin,
            'credentials_cookie_auth'
        )

        # also delete basic auth
        uf.manage_delObjects(['credentials_basic_auth'])

        # for some reason, we need to install the initial user...
        if not api.env.test_mode():
            try:
                uf.users.manage_addUser('admin', 'admin', 'admin', 'admin')
                uf.roles.assignRoleToPrincipal('Manager', 'admin')
            except KeyError:
                pass  # already a user

        if logger is not None:
            logger('Updated acl users')

    km = getattr(app, 'key_manager', None)
    if km is None:
        km = KeyManager()
        app.key_manager = km
        app._p_changed = 1
        if logger is not None:
            logger('adding key manager')

    sm = getGlobalSiteManager()
    sm.registerUtility(km, IKeyManager)



from plone.protect.interfaces import IDisableCSRFProtection
from plone.registry.interfaces import IRegistry
from zope import schema
from zope.component import queryUtility
from zope.interface import alsoProvides

import json


class SiteConfiguration:
    def __call__(self):
        form = self.request.form

        alsoProvides(self.request, IDisableCSRFProtection)
        self.request.response.setHeader('Content-type', 'application/json')

        registry = queryUtility(IRegistry)

        try:
            for field in form:
                if field in registry:

                    schema_field = registry.records[field]._field

                    fieldValue = form.get(field)
                    value = None

                    if isinstance(schema_field, schema.Bool):
                        #JSON doesn't want to convert them back to boolean
                        if fieldValue == 'true':
                            value = True
                        elif fieldValue == 'false':
                            value = False

                    elif isinstance(schema_field, schema.TextLine):
                        value = unicode(fieldValue)

                    elif isinstance(schema_field, schema.ASCII):
                        value = str(fieldValue)

                    else:
                        continue

                    registry[field] = value

            return json.dumps({
                'success': True,
                'start': self.context.absolute_url()
            })
        except:
            return json.dumps({
                'success': False,
                'errors': "There was a problem setting your configurations",
                'start': self.context.absolute_url()
            })



# -*- coding: utf-8 -*-
from castle.cms.browser.analytics import AnalyticsView
from castle.cms.social import COUNT_ANNOTATION_KEY
from castle.cms.testing import CASTLE_PLONE_INTEGRATION_TESTING
from plone import api
from plone.app.testing import login
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from zope.annotation.interfaces import IAnnotations

import json
import mock
import unittest


class TestAnalytics(unittest.TestCase):

    layer = CASTLE_PLONE_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer['portal']
        self.request = self.layer['request']

        login(self.portal, TEST_USER_NAME)
        setRoles(self.portal, TEST_USER_ID, ('Member', 'Manager'))
        if 'front-page' not in self.portal:
            api.content.create(type='Document', id='front-page',
                               container=self.portal)

            self.portal.setDefaultPage('front-page')
        self.page = self.portal['front-page']

        annotations = IAnnotations(self.page)
        annotations[COUNT_ANNOTATION_KEY] = {
            'twitter': 5,
            'facebook': 5,
            'linkedin': 5
        }

    def test_get_stats(self):
        view = AnalyticsView(self.page, self.request)
        self.assertEquals(json.loads(view())['data']['twitter'], 5)

    @mock.patch('castle.cms.cache.get')
    @mock.patch('castle.cms.cache.set')
    @mock.patch('castle.cms.services.google.get_ga_profile')
    @mock.patch('castle.cms.services.google.get_ga_service')
    def test_ga_api_call(self, get_ga_service, get_ga_profile, cache_set, cache_get):
        cache_get.side_effect = KeyError()
        view = AnalyticsView(self.page, self.request)
        self.request.form.update({
            'params': json.dumps({'foo': 'bar'})
        })
        view.ga_api_call(['/'])
        self.assertEquals(get_ga_service.call_count, 1)
        self.assertEquals(get_ga_profile.call_count, 1)

    @mock.patch('castle.cms.cache.get')
    @mock.patch('castle.cms.cache.set')
    @mock.patch('castle.cms.services.google.get_ga_profile')
    def test_get_ga_profile(self, get_ga_profile, cache_set, cache_get):
        cache_get.side_effect = KeyError()
        view = AnalyticsView(self.page, self.request)
        service_mock = mock.MagicMock()
        view.get_ga_profile(service_mock)
        self.assertEquals(cache_set.call_count, 1)
        self.assertEquals(cache_get.call_count, 1)
        self.assertEquals(get_ga_profile.call_count, 1)

    def test_get_paths(self):
        view = AnalyticsView(self.page, self.request)
        paths = view.get_paths()

        self.assertTrue('/' in paths)
        self.assertTrue('/front-page' in paths)



from Acquisition import aq_parent
from castle.cms.interfaces import ITileView
from plone.dexterity.interfaces import IDexterityContent
from Products.Five import BrowserView
from zope.component import getAdapters
from zope.globalrequest import getRequest
from zope.interface import implements
from zope.schema.interfaces import IContextSourceBinder
from zope.schema.vocabulary import SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary


def getTileViews(context, request, tile_name):
    adapters = getAdapters((context, request), ITileView)
    views = []
    for _, adapter in adapters:
        if adapter.tile_name == tile_name:
            views.append(adapter)
    return list(sorted(views, key=lambda x: x.order))


def getTileView(context, request, tile_name, view_name, default=None):
    for View in getTileViews(context, request, tile_name):
        if View.name == view_name:
            return View
    # use default maybe..
    if default is not None:
        return getTileView(context, request, tile_name, default)


class BaseTileView(BrowserView):
    implements(ITileView)
    order = 0
    name = None
    tile_name = None
    label = None
    preview = None
    hidden = False

    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.tile = None

    def __call__(self):
        return self.index()


class TileViewsSource(object):
    implements(IContextSourceBinder)

    def __init__(self, tile_name):
        self.tile_name = tile_name

    def __call__(self, context):
        terms = []
        if not IDexterityContent.providedBy(context):
            context = aq_parent(context)
        for View in getTileViews(context, getRequest(), self.tile_name):
            if View.hidden:
                continue
            label = View.label
            if label is None:
                label = View.name.capitalize()
            terms.append(SimpleTerm(View.name, View.name, label))
        return SimpleVocabulary(terms)



from boto.s3.connection import ProtocolIndependentOrdinaryCallingFormat
from boto.s3.connection import S3Connection
from datetime import datetime
from persistent.mapping import PersistentMapping
from plone.namedfile.file import NamedBlobFile
from plone.registry.interfaces import IRegistry
from plone.uuid.interfaces import IUUID
from time import time
from urlparse import urlparse
from zope.annotation.interfaces import IAnnotations
from zope.component import getUtility

import io
import logging
import math
import urllib


logger = logging.getLogger('castle.cms')


CHUNK_SIZE = 12428800
STORAGE_KEY = 'castle.cms.aws'
FILENAME = u'moved-to-s3'
EXPIRES_IN = 30 * 24 * 60 * 60

# define prefix for keys on where to store data in aws
KEY_PREFIX = 'files/'


def get_bucket(s3_bucket=None):
    registry = getUtility(IRegistry)
    s3_id = registry.get('castle.aws_s3_key', None)
    s3_key = registry.get('castle.aws_s3_secret', None)
    if s3_bucket is None:
        s3_bucket = registry.get('castle.aws_s3_bucket_name', None)

    if not s3_id or not s3_key or not s3_bucket:
        return None, None

    endpoint = registry.get('castle.aws_s3_host_endpoint', None)
    if endpoint:
        s3_conn = S3Connection(
            s3_id, s3_key, host=endpoint,
            calling_format=ProtocolIndependentOrdinaryCallingFormat())
    else:
        s3_conn = S3Connection(s3_id, s3_key)

    return s3_conn, s3_conn.get_bucket(s3_bucket)


def move_file(obj):
    _, bucket = get_bucket()
    if bucket is None:
        return

    uid = IUUID(obj)
    if not uid:
        logger.info('Could not get uid of object')
        return

    key = KEY_PREFIX + uid
    filename = obj.file.filename
    if not isinstance(filename, unicode):
        filename = unicode(filename, 'utf-8', errors="ignore")
    filename = urllib.quote(filename.encode("utf8"))
    disposition = "attachment; filename*=UTF-8''%s" % filename

    size = obj.file.getSize()
    chunk_count = int(math.ceil(size / float(CHUNK_SIZE)))
    content_type = obj.file.contentType
    mp = bucket.initiate_multipart_upload(key, metadata={
        'Content-Type': content_type,
        'Content-Disposition': disposition
    })

    blob_fi = obj.file._blob.open('r')

    for i in range(chunk_count):
        chunk = blob_fi.read(CHUNK_SIZE)
        fp = io.BytesIO(chunk)
        mp.upload_part_from_file(fp, part_num=i + 1)

    mp.complete_upload()
    blob_fi.close()

    obj._p_jar.sync()
    obj.file = NamedBlobFile(data='', contentType=obj.file.contentType, filename=FILENAME)
    obj.file.original_filename = filename
    obj.file.original_content_type = content_type
    obj.file.original_size = size

    set_permission(obj)


def set_permission(obj):
    s3_conn, bucket = get_bucket()
    if bucket is None:
        return
    uid = IUUID(obj)
    key_name = KEY_PREFIX + uid
    key = bucket.get_key(key_name)
    if key is None:
        return

    perms = obj.rolesOfPermission("View")
    public = False
    for perm in perms:
        if perm["name"] == "Anonymous":
            if perm["selected"] != "":
                public = True
                break

    annotations = IAnnotations(obj)
    info = annotations.get(STORAGE_KEY, PersistentMapping())
    if public:
        key.make_public()
        expires_in = 0
        url = 'https://{host}/{bucket}/{key}'.format(
            host=s3_conn.server_name(),
            bucket=bucket.name,
            key=key_name)
    else:
        key.set_canned_acl('private')
        url = key.generate_url(EXPIRES_IN)
        expires_in = EXPIRES_IN

    info.update({
        'url': url,
        'expires_in': expires_in,
        'generated_on': time()
    })
    annotations[STORAGE_KEY] = info


def delete_file(uid):
    _, bucket = get_bucket()
    if bucket is None:
        return

    key_name = KEY_PREFIX + uid
    key = bucket.get_key(key_name)
    if key:
        bucket.delete_key(key_name)


def uploaded(obj):
    annotations = IAnnotations(obj)
    info = annotations.get(STORAGE_KEY, PersistentMapping())
    return 'url' in info


def _one(val):
    """Wonky, values returning tuple sometimes?"""
    if type(val) in (tuple, list, set):
        if len(val) > 0:
            val = val[0]
        else:
            return
    return val


def swap_url(url, registry=None, base_url=None):
    """
    aws url: https://s3-us-gov-west-1.amazonaws.com/bucketname/archives/path/to/resource
    base url: http://foo.com/
    swapped: http://foo.com/archives/path/to/resource
    """
    if base_url is None:
        if registry is None:
            registry = getUtility(IRegistry)
        base_url = registry.get('castle.aws_s3_base_url', None)

    if base_url:
        parsed_url = urlparse(url)
        url = '{}/{}'.format(base_url.strip('/'), '/'.join(parsed_url.path.split('/')[2:]))
        if parsed_url.query:
            url += '?' + parsed_url.query
    return url


def get_url(obj):
    annotations = IAnnotations(obj)
    info = annotations.get(STORAGE_KEY, PersistentMapping())

    if info['expires_in'] != 0:
        # not public, check validity of url
        generated_on = info['generated_on']
        expires = datetime.fromtimestamp(generated_on + info['expires_in'])
        if datetime.utcnow() > expires:
            # need a new url
            _, bucket = get_bucket()
            uid = IUUID(obj)
            key = bucket.get_key(KEY_PREFIX + uid)
            url = key.generate_url(EXPIRES_IN)
            info.update({
                'url': url,
                'generated_on': time(),
                'expires_in': EXPIRES_IN
            })
            annotations[STORAGE_KEY] = info
    url = _one(info['url'])
    if not url.startswith('http'):
        url = 'https:' + url

    return swap_url(url)



# login/lockout and shield integration
from AccessControl.AuthEncoding import _randrange
from AccessControl.AuthEncoding import constant_time_compare
from AccessControl.AuthEncoding import pw_encrypt
from AccessControl.AuthEncoding import registerScheme
from AccessControl.AuthEncoding import SSHADigestScheme
from binascii import a2b_base64
from binascii import b2a_base64
from castle.cms import shield
from castle.cms.interfaces import ICastleLayer
from castle.cms.lockout import LOGGED_IN_MARKER_KEY
from castle.cms.lockout import SessionManager
from castle.cms.utils import get_context_from_request
from plone import api
from plone.uuid.interfaces import IUUID
from Products.PluggableAuthService.interfaces.events import IUserLoggedInEvent
from Products.PluggableAuthService.interfaces.events import IUserLoggedOutEvent
from urlparse import urlparse
from zExceptions import Redirect
from zope.component import adapter
from zope.component.hooks import getSite
from zope.globalrequest import getRequest
from ZPublisher.interfaces import IPubAfterTraversal
from ZPublisher.interfaces import IPubBeforeCommit

import binascii
import logging


try:
    import argon2
except ImportError:
    argon2 = None


log = logging.getLogger(__name__)


@adapter(IUserLoggedOutEvent)
def onUserLogsOut(event):
    request = getRequest()
    if request is None:
        return
    site = api.portal.get()
    user = api.user.get_current()
    try:
        session_manager = SessionManager(site, request, user)
        resp = request.response
        resp.expireCookie(session_manager.cookie_name)
        session_manager.delete()
    except:
        pass


@adapter(IUserLoggedInEvent)
def onUserLogsIn(event):
    """
    let us know that the user was logged in here successfully
    """
    request = getRequest()
    if request is None:
        return
    if not ICastleLayer.providedBy(request):
        return
    request.environ[LOGGED_IN_MARKER_KEY] = 'yes'

    # do not even allow user to login if the account has been
    # disabled
    if api.user.get_roles(user=event.object) == ['Authenticated']:
        site = getSite()
        raise Redirect('%s/@@disabled-user' % site.absolute_url())


@adapter(IPubAfterTraversal)
def afterTraversal(event):
    """
    check it should be blocked by lockout
    """
    request = event.request
    if not ICastleLayer.providedBy(request):
        return

    shield.protect(request)

    resp = request.response

    context = get_context_from_request(request)
    cache_tags = set([
        getattr(context, 'portal_type', '').lower().replace(' ', '-'),
        getattr(context, 'meta_type', '').lower().replace(' ', '-'),
        IUUID(context, ''),
        urlparse(request.URL).netloc.lower().replace('.', '').replace(':', '')
    ])

    resp.setHeader('Cache-Tag', ','.join(t for t in cache_tags if t))

    # Prevent IE and Chrome from incorrectly detecting non-scripts as scripts
    resp.setHeader('X-Content-Type-Options', 'nosniff')
    # prevent some XSS from browser
    resp.setHeader('X-XSS-Protection', '1; mode=block')


@adapter(IPubBeforeCommit)
def beforeCommit(event):
    """
    Couple causes here:

    1. Lockout support
        check if user attempted to login to the site.
        If success, reset counter, if fail, tally it.

    """
    request = event.request

    if not ICastleLayer.providedBy(request):
        return

    site = api.portal.get()

    resp = request.response
    contentType = resp.getHeader('Content-Type')
    if site is None or contentType is None or not contentType.startswith('text/html'):
        return None

    # now, check user roles. If they have none, make sure to
    # throw an exception with message saying the user's account
    # is disabled
    user = api.user.get_current()
    if user.getId() is None:
        return
    if api.user.get_roles(user=user) == ['Authenticated']:
        # clear login cookies
        mt = api.portal.get_tool('portal_membership')
        mt.logoutUser(request)
        resp.redirect('%s/@@disabled-user' % site.absolute_url())

    session_manager = SessionManager(site, request, user)
    if not session_manager.has_session_id():
        # register new session with new id and storage
        session_manager.register()
    else:
        session = session_manager.get()
        if not session:
            session_manager.log({})
        else:
            if session_manager.expired(session):
                mt = api.portal.get_tool('portal_membership')
                mt.logoutUser(request)
                resp.expireCookie(session_manager.cookie_name)
                session_manager.delete()
                resp.redirect('%s/@@session-removed' % site.absolute_url())
            else:
                session_manager.log(session)


if argon2 is not None:
    # only available when installed...

    class Argon2Scheme(SSHADigestScheme):
        """
        scheme follows this format...

        argon:salt:hashed
        """
        ph = argon2.PasswordHasher()

        def _format(self, pw):
            return b2a_base64(self.ph.hash(pw))

        def encrypt(self, pw):
            pw = str(pw)
            return self._format(pw)

        def validate(self, reference, attempt):
            try:
                ref = a2b_base64(reference)
            except binascii.Error:
                # Not valid base64.
                return False

            try:
                return self.ph.verify(ref, attempt)
            except argon2.exceptions.Argon2Error:
                return False

    registerScheme('argon2', Argon2Scheme())
    # we patch this to be default because plone doesn't provide a param
    pw_encrypt.func_defaults = ('argon2',)



from castle.cms.events import IMetaTileEditedEvent
from castle.cms.interfaces import ITrashed
from castle.cms.utils import ESConnectionFactoryFactory
from datetime import datetime
from elasticsearch import TransportError
from plone import api
from plone.app.iterate.interfaces import IAfterCheckinEvent
from plone.app.iterate.interfaces import ICancelCheckoutEvent
from plone.app.iterate.interfaces import ICheckoutEvent
from plone.app.iterate.interfaces import IWorkingCopyDeletedEvent
from plone.registry.interfaces import IRegistry
from plone.uuid.interfaces import IUUID
from Products.DCWorkflow.interfaces import IAfterTransitionEvent
from Products.PluggableAuthService.interfaces.events import ICredentialsUpdatedEvent
from Products.PluggableAuthService.interfaces.events import IPrincipalCreatedEvent
from Products.PluggableAuthService.interfaces.events import IPrincipalDeletedEvent
from Products.PluggableAuthService.interfaces.events import IPropertiesUpdatedEvent
from Products.PluggableAuthService.interfaces.events import IUserLoggedInEvent
from Products.PluggableAuthService.interfaces.events import IUserLoggedOutEvent
from zope.component import ComponentLookupError
from zope.component import getUtility
from zope.globalrequest import getRequest
from zope.interface import providedBy
from zope.lifecycleevent.interfaces import IObjectAddedEvent
from zope.lifecycleevent.interfaces import IObjectCopiedEvent
from zope.lifecycleevent.interfaces import IObjectModifiedEvent
from zope.lifecycleevent.interfaces import IObjectMovedEvent
from zope.lifecycleevent.interfaces import IObjectRemovedEvent

import threading
import transaction


class DefaultRecorder(object):
    valid = True

    def __init__(self, data, event, obj=None):
        self.data = data
        self.event = event
        self.obj = obj

    def __call__(self):
        data = {
            'type': self.data._type,
            'name': self.data.name,
            'summary': self.data.summary,
            'user': api.user.get_current().getId(),
            'date': datetime.utcnow().isoformat(),
            'request_uri': getRequest().URL
        }
        if self.obj is not None:
            data['object'] = IUUID(self.obj)
            data['path'] = '/'.join(self.obj.getPhysicalPath())
        return data


class LoggedInRecorder(DefaultRecorder):
    def __call__(self):
        data = super(LoggedInRecorder, self).__call__()
        data['user'] = self.event.object.getId()
        return data


class WorkflowRecorder(DefaultRecorder):

    @property
    def valid(self):
        return self.event.transition is not None

    def __call__(self):
        data = super(WorkflowRecorder, self).__call__()
        data['summary'] = 'previous: %s, new: %s, transition: %s' % (
            self.event.old_state.title,
            self.event.new_state.title,
            self.event.transition.title,
        )
        return data


class MetaTileRecorder(DefaultRecorder):
    def __call__(self):
        data = super(MetaTileRecorder, self).__call__()
        data['summary'] = 'Slot edited: {}'.format(self.event.tile_id)
        return data


class AuditData(object):

    def __init__(self, _type, name, summary=None, recorder_class=DefaultRecorder):
        self._type = _type
        self.name = name
        self.summary = summary
        self.recorder_class = recorder_class

    def get_recorder(self, event, obj=None):
        return self.recorder_class(self, event, obj)


_registered = {
    IAfterCheckinEvent: AuditData('working copy support', 'Check in'),
    ICheckoutEvent: AuditData('working copy support', 'Check out'),
    ICancelCheckoutEvent: AuditData('working copy support', 'Cancel checkout'),
    IWorkingCopyDeletedEvent: AuditData('working copy support', 'Working copy deleted'),
    IObjectAddedEvent: AuditData('content', 'Created'),
    IObjectCopiedEvent: AuditData('content', 'Copied'),
    IObjectModifiedEvent: AuditData('content', 'Modified'),
    IObjectMovedEvent: AuditData('content', 'Moved'),
    IObjectRemovedEvent: AuditData('content', 'Deleted'),
    IPrincipalCreatedEvent: AuditData('user', 'Created'),
    ITrashed: AuditData('content', 'Trashed'),
    IUserLoggedInEvent: AuditData('user', 'Logged in', recorder_class=LoggedInRecorder),
    IUserLoggedOutEvent: AuditData('user', 'Logged out'),
    IPrincipalDeletedEvent: AuditData('user', 'Deleted'),
    ICredentialsUpdatedEvent: AuditData('user', 'Password updated'),
    IPropertiesUpdatedEvent: AuditData('user', 'Properties updated'),
    IAfterTransitionEvent: AuditData('workflow', 'Transition', recorder_class=WorkflowRecorder),
    IMetaTileEditedEvent: AuditData('slots', 'Slot edited', recorder_class=MetaTileRecorder),
}


es_doc_type = 'entry'


def _create_index(es, index_name):
    mapping = {'properties': {
        'type': {'store': False, 'type': 'string', 'index': 'not_analyzed'},
        'name': {'store': False, 'type': 'string', 'index': 'not_analyzed'},
        'summary': {'store': False, 'type': 'string', 'index': 'not_analyzed'},
        'user': {'store': False, 'type': 'string', 'index': 'not_analyzed'},
        'request_uri': {'store': False, 'type': 'string', 'index': 'not_analyzed'},
        'date': {'store': False, 'type': 'date'},
        'object': {'store': False, 'type': 'string', 'index': 'not_analyzed'},
        'path': {'store': False, 'type': 'string', 'index': 'not_analyzed'},
    }}
    if not es.indices.exists(index_name):
        es.indices.create(index_name)
    es.indices.put_mapping(
        doc_type=es_doc_type,
        body=mapping,
        index=index_name)


def get_index_name(site_path=None):
    if site_path is None:
        site_path = '/'.join(api.portal.get().getPhysicalPath())
    return site_path.replace('/', '').replace(' ', '').lower()


def _record(conn_factory, site_path, data):
    index_name = get_index_name(site_path)
    es = conn_factory()
    try:
        es.index(index=index_name, doc_type=es_doc_type, body=data)
    except TransportError as ex:
        if 'InvalidIndexNameException' in ex.error:
            try:
                _create_index(es, index_name)
            except TransportError:
                return
            _record(conn_factory, site_path, data)


def record(success, recorder, site_path, conn):
    if not success:
        return
    if recorder.valid:
        data = recorder()
        thread = threading.Thread(target=_record, args=(
            conn,
            site_path,
            data
        ))
        thread.start()


def event(obj, event=None):

    if event is None:
        # some events don't include object
        event = obj
        obj = None

    iface = providedBy(event).declared[0]
    if iface in _registered:
        try:
            registry = getUtility(IRegistry)
        except ComponentLookupError:
            return
        if not registry.get('collective.elasticsearch.interfaces.IElasticSettings.enabled', False):
            return

        if obj is not None and ITrashed.providedBy(obj):
            # special handling for ITrashed...
            # we don't want to record transition events for trashing
            # and we want a special psuedo trash event
            # then, we still want an event when it gets deleted for good
            if iface not in (IAfterTransitionEvent, IObjectRemovedEvent):
                # dive out if not IAfterTransitionEvent or object removed event
                return
            if iface == IObjectRemovedEvent:
                audit_data = _registered[iface]
            else:
                audit_data = _registered[ITrashed]
        else:
            audit_data = _registered[iface]

        recorder = audit_data.get_recorder(event, obj)
        site_path = '/'.join(api.portal.get().getPhysicalPath())
        transaction.get().addAfterCommitHook(record, args=(
            recorder, site_path, ESConnectionFactoryFactory(registry)))
    else:
        pass



from castle.cms import tasks
from castle.cms.interfaces import ITrashed
from plone import api
from plone.dexterity.interfaces import IDexterityContainer
from zope.interface import alsoProvides
from zope.interface import noLongerProvides


def object(ob):
    state = api.content.get_state(obj=ob, default=None)
    if state is not None and state != 'private':
        try:
            api.content.transition(obj=ob, to_state='private')
        except:
            pass

    alsoProvides(ob, ITrashed)

    ob.setModificationDate()
    ob.reindexObject(idxs=['trashed', 'object_provides', 'modified'])
    if IDexterityContainer.providedBy(ob):
        tasks.trash_tree.delay(ob)


def restore(ob):
    noLongerProvides(ob, ITrashed)
    ob.reindexObject(idxs=['trashed', 'object_provides'])
    if IDexterityContainer.providedBy(ob):
        tasks.trash_tree.delay(ob)



from castle.cms.tiles.base import BaseTile
from plone.supermodel import model
from zope import schema
from zope.schema.vocabulary import SimpleVocabulary

import re


class FacebookPageTile(BaseTile):
    def render(self):
        content = self.data.get('href', '') or ''

        urlRegEx = ".*facebook\.com/([a-zA-Z0-9]+)/?"
        urlParser = re.compile(urlRegEx)
        matches = urlParser.match(content)

        if matches is None:
            return ''

        self.pageID = matches.group(1)

        self.parameters = {}
        validFields = IFacebookPageTileSchema._InterfaceClass__attrs.keys()

        # Prevent any additional values from being passed to FB
        for key in self.data.keys():
            if key in validFields:
                self.parameters[key] = self.data.get(key, '')

        return self.index()


class IFacebookPageTileSchema(model.Schema):
    href = schema.URI(
        title=u"URL",
        description=u"URL to the Facebook page to include in the tile."
    )

    hide_cover = schema.Bool(
        title=u"Hide cover photo",
        description=u"Don't include the page's cover photo in the tile."
    )

    show_facepile = schema.Bool(
        title=u"Show friend's faces",
        description=u"Show friends of the current user who liked the page."
    )

    small_header = schema.Bool(
        title=u"Use small header",
        description=u"Show a smaller, less detailed header."
    )

    timeline = schema.Choice(
        title=u"Show timeline",
        description=u"Show past posts by the Facebook page being embedded.",
        vocabulary=SimpleVocabulary.fromValues(['', 'timeline', 'events'])
    )



import base64
import re

from castle.cms.behaviors.location import ILocation
from plone.app.blocks.layoutbehavior import ILayoutAware
from plone.app.contenttypes.behaviors.richtext import IRichText
from plone.app.dexterity.behaviors.metadata import IBasic
from plone.app.dexterity.behaviors.metadata import ICategorization
from plone.app.dexterity.behaviors.metadata import IPublication
from plone.app.textfield.value import RichTextValue
from plone.namedfile.file import NamedBlobFile
from plone.namedfile.file import NamedBlobImage


FOLDER_DEFAULT_PAGE_LAYOUT = u"""
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
                      "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en"
      data-layout="./@@page-site-layout">
  <body>
    <div data-panel="content">
      <div class="mosaic-grid-row"
           data-grid='{"type": "row"}'>
        <div class="mosaic-grid-cell mosaic-width-full mosaic-position-leftmost"
             data-grid='{"type": "cell", "info":{"xs": "true", "sm": "true", "lg": "true",
                         "pos": {"x": 1, "width": 12}}}'>
          <div class="movable removable mosaic-tile mosaic-IDublinCore-title-tile">
          <div class="mosaic-tile-content">
          <div data-tile="./@@plone.app.standardtiles.field?field=IDublinCore-title"></div>
          </div>
          </div>
          <div class="movable removable mosaic-tile mosaic-IDublinCore-description-tile">
          <div class="mosaic-tile-content">
          <div data-tile="./@@plone.app.standardtiles.field?field=IDublinCore-description"></div>
          </div>
          </div>
          <div class="movable removable mosaic-tile mosaic-text-tile">
          <div class="mosaic-tile-content">
            %s
          </div>
          </div>
        </div>
      </div>
    </div>
  </body>
</html>
"""
_types = {}


def isBase64(s):
    return (len(s) % 4 == 0) and re.match('^[A-Za-z0-9+/]+[=]{0,2}$', s)


def decodeFileData(data):
    if isinstance(data, OFS.Image.Image):
        data = data.data
        if isinstance(data, OFS.Image.Pdata):
            return str(data)
        return data
    if isBase64(data):
        return base64.b64decode(data)
    return data


def toUnicode(s):
    if not isinstance(s, unicode):
        s = s.decode('utf8')
    return s


def _generate_file_repo_object_id(path):
    path = path.lower()
    if path[0] == '/':
        path = path[1:]
    for _type in ('file', 'audio', 'video', 'videos', 'image', 'images'):
        path = path.replace('%s-repository/' % _type, '').replace('/' + _type, '')
    return path.replace('/', '-')


class BaseImportType(object):
    layout = ''

    def __init__(self, data, path, read_phase):
        self.data = data
        self.field_data = data['data']
        self.path = path
        self.read_phase = read_phase
        self.original_type = data['portal_type']
        if data.get('is_default_page'):
            self.path = '/'.join(path.split('/')[:-1])
            if data.get('has_sibling_pages'):
                self.data['portal_type'] = 'Folder'

    def get_path(self):
        return self.path

    def get_data(self):
        data = {
            '_plone.uuid': self.data['uid']
        }
        if self.field_data.get('image'):
            im_data = self.field_data.get('image')
            if hasattr(im_data, 'read'):
                im_data = im_data.read()
            filename = self.field_data.get('image_filename')
            if not filename:
                filename = self.field_data['id']
            data['image'] = NamedBlobImage(
                data=decodeFileData(im_data),
                filename=toUnicode(filename))
            if not data['image'].contentType:
                data['image'].contentType = 'image/jpeg'
            data['imageCaption'] = self.field_data.get('imageCaption')
        return dict(
            id=self.field_data['id'],
            type=self.data['portal_type'],
            title=self.field_data['title'],
            description=self.field_data['description'],
            **data)

    def post_creation(self, obj):
        field_data = self.field_data
        bdata = ILayoutAware(obj, None)
        if bdata:
            bdata.contentLayout = '++contentlayout++default/document.html'
        bdata = IRichText(obj, None)
        if bdata:
            bdata.text = RichTextValue(field_data['text'], 'text/html', 'text/html')

        bdata = IBasic(obj, None)
        if bdata:
            bdata.title = field_data['title']
            bdata.description = field_data['description']
        else:
            obj.title = field_data['title']
            obj.description = field_data['description']

        bdata = ICategorization(obj, None)
        if bdata:
            bdata.subjects = field_data['subject']

            bdata = IPublication(obj)
            if field_data['effectiveDate']:
                bdata.effective = field_data['effectiveDate'].asdatetime()

        ldata = ILocation(obj, None)
        if ldata:
            if field_data.get('location'):
                ldata.locations = [field_data['location']]

            if field_data.get('newsLocation'):
                if ldata.locations:
                    ldata.locations.append(field_data['newsLocation'])
                else:
                    ldata.locations = [field_data['newsLocation']]

        obj.modification_date = field_data['modification_date']
        obj.creation_date = field_data['creation_date']

        bdata = ILayoutAware(obj, None)
        if bdata:
            if self.data['portal_type'] == 'Folder' and 'text' in self.field_data:
                bdata.content = FOLDER_DEFAULT_PAGE_LAYOUT % self.field_data['text']
            elif self.layout:
                bdata.contentLayout = self.layout


class DocumentType(BaseImportType):
    layout = '++contentlayout++default/document.html'
_types['Document'] = DocumentType

class FolderType(BaseImportType):
    layout = '++contentlayout++castle/folder.html'
_types['Folder'] = FolderType


class NewsItemType(BaseImportType):

    layout = '++contentlayout++castle/newsitem.html'

_types['News Item'] = NewsItemType


class FileType(BaseImportType):
    layout = None

    def __init__(self, data, path, read_phase):
        super(FileType, self).__init__(data, path, read_phase)
        filename = self.field_data.get('file_filename')
        _, ext = filename.lower().rsplit('.', 1)
        _id = _generate_file_repo_object_id(path)
        self.field_data['id'] = _id
        if ext in ('wav', 'mp3'):
            self.data['portal_type'] = 'Audio'
            self.path = '/audio-repository/' + _id
        elif ext in ['webm', 'ogv', 'avi', 'wmv', 'm4v',
                     'mpg', 'mpeg', 'flv', 'mp4', 'mov']:
            self.data['portal_type'] = 'Video'
            self.path = '/video-repository/' + _id
        else:
            self.path = '/file-repository/' + _id

    def get_data(self):
        data = super(FileType, self).get_data()
        data['file'] = NamedBlobFile(
            data=decodeFileData(self.field_data['file'].read()),
            filename=toUnicode(self.field_data.get('file_filename')))
        return data
_types['File'] = FileType


class ImageType(BaseImportType):

    def __init__(self, data, path, read_phase):
        super(ImageType, self).__init__(data, path, read_phase)
        _id = _generate_file_repo_object_id(path)
        self.field_data['id'] = _id
        self.path = '/image-repository/' + _id
_types['Image'] = ImageType


class LinkType(BaseImportType):

    youtube = False

    def __init__(self, data, path, read_phase):
        super(LinkType, self).__init__(data, path, read_phase)
        url = self.field_data['remoteUrl']
        if 'youtube' in url:
            self.youtube = True
            self.data['portal_type'] = 'Video'
            _id = _generate_file_repo_object_id(path)
            self.path = '/video-repository/' + _id

    def get_path(self):
        if self.data['portal_type'] == 'Video':
            return '/video-repository/' + self.field_data['id']
        return super(LinkType, self).get_path()

    def get_data(self):
        data = super(LinkType, self).get_data()
        if self.youtube:
            data['youtube_url'] = self.field_data['remoteUrl']
        else:
            data['remoteUrl'] = self.field_data['remoteUrl']
        return data
_types['Link'] = LinkType


def getImportType(data, path, read_phase):
    if data['portal_type'] in _types:
        return _types[data['portal_type']](data, path, read_phase)
    return BaseImportType(data, path, read_phase)



from zope.component import getUtility
from plone import api
from Acquisition import aq_inner
from plone.registry.interfaces import IRegistry
from plone.uuid.interfaces import IUUID
from plone.app.standardtiles import discussion

DISQUS_CODE = """
<div id="disqus_thread"></div>
<script>
  var disqus_config = function () {
    this.page.url = "%(url)s";
    this.page.identifier = "%(uid)s";
  };
  (function() {
    var d = document, s = d.createElement('script');
    s.src = '//%(shortname)s.disqus.com/embed.js';

    s.setAttribute('data-timestamp', +new Date());
    (d.head || d.body).appendChild(s);
  })();
</script>
"""


class DiscussionTile(discussion.DiscussionTile):

    def __call__(self):
        context = aq_inner(self.context)
        convo = context.restrictedTraverse('@@conversation_view', None)
        if not convo or not convo.enabled():
            return u'<html></html>'

        registry = getUtility(IRegistry)
        try:
            disqus_shortname = registry['plone.disquis_shortname']
        except:
            disqus_shortname = None

        if not disqus_shortname:
            return super(DiscussionTile, self).__call__()

        url = self.context.absolute_url()
        public_url = registry.get('plone.public_url', None)
        if public_url:
            url = url.replace(
                api.portal.get().absolute_url(),
                public_url.rstrip('/') + '/')

        return DISQUS_CODE % {
            'url': url,
            'uid': IUUID(self.context),
            'shortname': disqus_shortname
        }
